# coding: utf-8
# Distributed under the terms of the MIT License.

""" This script mimics the dispersion.pl script bundled
with CASTEP. For the given bands file, a bandstructure
is created. If a <seed>.adaptive.dat file exists
then a combined BS/DOS plot will be created.
"""

from os.path import isfile
import argparse
import glob
from matador import __version__, script_epilog


def main():
    """Parse args and run the script."""
    parser = argparse.ArgumentParser(
        prog="dispersion",
        epilog=script_epilog,
        description="simple plotting script for bandstructures/DOS based on matador",
        add_help=True,
    )

    parser.add_argument(
        "--version", action="version", version="matador version " + __version__ + "."
    )
    parser.add_argument(
        "--pdf", action="store_true", help="save pdf rather than showing plot in X"
    )
    parser.add_argument(
        "--png", action="store_true", help="save png rather than showing plot in X"
    )
    parser.add_argument(
        "--svg", action="store_true", help="save svg rather than showing plot in X"
    )
    parser.add_argument(
        "--labels", type=str, nargs="*", help="list of legend labels, comma separated"
    )
    parser.add_argument("--dos_only", action="store_true", help="only plot DOS")
    parser.add_argument("--bs_only", action="store_true", help="only plot dispersion")
    parser.add_argument(
        "--preserve_kspace_distance",
        action="store_true",
        help="when linearising kpoint path, ensure distance in reciprocal space is conserved",
    )
    parser.add_argument("--cmap", type=str, help="matplotlib colourmap name to use")
    parser.add_argument(
        "--spin_only", type=str, help='either "up" or "down" to only plot one channel'
    )
    parser.add_argument(
        "-interp",
        "--pdis_interpolation_factor",
        type=float,
        help="multiple by which to interpolate pDIS bands (DEFAULT: 2)",
    )
    parser.add_argument(
        "-scale",
        "--pdis_point_scale",
        type=float,
        help="point scale in pDIS plots (DEFAULT: 25)",
    )
    parser.add_argument(
        "-p",
        "--projectors_to_plot",
        type=str,
        help=(
            "a list of projectors to plot, in the format element:orbital, "
            "e.g. -p K:s,P:p. If the orbital is omitted, all orbitals will be used for that element."
        ),
    )
    parser.add_argument(
        "--unstacked_pdos",
        action="store_true",
        help="plot PDOS as overlap rather than stack",
    )
    parser.add_argument(
        "--no_pdis",
        action="store_true",
        help="do not try to plot PDIS, even if its available",
    )
    parser.add_argument(
        "--band_reorder",
        action="store_true",
        help="try to reorder bands based on local gradients",
    )
    parser.add_argument(
        "--pdos_hide_sum", action="store_true", help="plot PDOS without sum pDOS"
    )
    parser.add_argument(
        "--projector_colours",
        type=str,
        nargs="+",
        help="override all projector colour options with a list of matplotlib-interpretable colours, "
        "e.g. --projector_colours red blue #ff00ff",
    )
    parser.add_argument(
        "--colours",
        type=str,
        nargs="+",
        help="override all colour options with a list of matplotlib-interpretable colours, "
        "e.g. --colours red blue #ff00ff",
    )
    parser.add_argument(
        "-g", "--gap", action="store_true", help="plot position and size of band gap"
    )
    parser.add_argument(
        "-ph",
        "--phonons",
        action="store_true",
        default=False,
        help="plot phonon calculation, rather than electronic",
    )
    parser.add_argument(
        "-ir",
        "--infrared",
        action="store_true",
        default=False,
        help="plot infrared spectrum from file",
    )
    parser.add_argument(
        "-ir_ss",
        "--infrared_step_size",
        type=float,
        nargs="?",
        default=1.0,
        help="step size in cm^{-1} on x-axis for IR plots; must be > 0, default is 1",
    )
    parser.add_argument(
        "-gw",
        "--gaussian_width",
        type=float,
        help=(
            "smearing width for DOS from .bands_dos (default: 0.1 eV) or "
            ".phonon_dos files (default: 10 1/cm)"
        ),
    )
    parser.add_argument(
        "--highlight_bands",
        nargs="+",
        type=int,
        help="specify band numbres to highlight in plot",
    )
    parser.add_argument(
        "-v", "--verbosity", type=int, default=0, help="control verbosity of output"
    )
    parser.add_argument(
        "-figsize",
        "--figsize",
        nargs="+",
        type=int,
        help="figure size in inches to pass to matplotlib",
    )
    parser.add_argument(
        "-pw",
        "--plot_window",
        nargs="+",
        type=float,
        help="energy window [x, y] or [-x, x] to plot either side of E_F (eV)\
                             (DEFAULT: 5 eV)",
    )
    parser.add_argument(
        "seed",
        type=str,
        nargs="+",
        help="seedname or related filename (e.g. bands or dos file)",
    )
    kwargs = vars(parser.parse_args())
    if "gap" in kwargs:
        gap = kwargs["gap"]
        del kwargs["gap"]

    seeds = kwargs.get("seed")
    exts_to_strip = [
        "bands",
        "linear.dat",
        "adaptive.dat",
        "pdis.dat",
        "bands_dos",
        "pdos.dat",
        "phonon",
        "phonon_dos",
    ]

    for ind, seed in enumerate(seeds):
        for ext in exts_to_strip:
            seeds[ind] = seeds[ind].replace("." + ext, "")

    verbosity = kwargs.get("verbosity")
    phonons = kwargs.get("phonons")
    ir = kwargs.get("infrared")
    if kwargs.get("labels"):
        labels = " ".join(kwargs.get("labels")).split(",")
    else:
        labels = None

    if kwargs["no_pdis"]:
        plot_pdis = False
    else:
        plot_pdis = True
    del kwargs["no_pdis"]

    cmap = kwargs.get("cmap")
    band_reorder = kwargs.get("band_reorder", False)

    from matador.plotting import plot_spectral, plot_ir_spectrum

    del kwargs["seed"]
    del kwargs["verbosity"]
    del kwargs["labels"]
    del kwargs["cmap"]
    del kwargs["band_reorder"]

    bandstructure = False
    dos = False

    for seed in seeds:
        if not phonons and not ir:
            bs_seed = seed + ".bands"
            bandstructure = isfile(bs_seed)
            dos_seeds = glob.glob(seed + "*.dat")
            if isfile(seed + ".bands_dos"):
                dos_seeds.append(seed + ".bands_dos")
            dos = any([isfile(dos_seed) for dos_seed in dos_seeds])

        elif phonons and not ir:
            phonon_seed = seed + ".phonon"
            bandstructure = isfile(phonon_seed)
            dos_seed = seed + ".phonon_dos"
            dos = isfile(dos_seed)

        elif ir:
            if len(seeds) > 1:
                exit("Multiple seeds not supported for IR plot.")
            ir_seed = seed + ".phonon"

    if bandstructure:
        cell_seed = seed + ".cell"
        cell = isfile(cell_seed)
    else:
        cell = False

    if kwargs.get("dos_only") and dos:
        bandstructure = False

    if kwargs.get("bs_only") and bandstructure:
        dos = False

    if not any([bandstructure, dos, ir]):
        raise SystemExit("Unable to find files for seedname {}".format(seed))

    if not ir:
        return plot_spectral(
            seeds,
            plot_bandstructure=bandstructure,
            plot_dos=dos,
            plot_pdis=plot_pdis,
            cell=cell,
            gap=gap,
            verbosity=verbosity,
            labels=labels,
            cmap=cmap,
            band_reorder=band_reorder,
            **kwargs
        )

    if ir:
        return plot_ir_spectrum(
            ir_seed, bin_width=kwargs["infrared_step_size"], **kwargs
        )

    exit("Issue plotting {}: did you specify -ph/-ir appropriately?".format(seeds))


if __name__ == "__main__":
    main()



""" This submodule contains aliases for some useful constants. """

from scipy.constants import physical_constants

FARADAY_CONSTANT_Cpermol = physical_constants["Faraday constant"][0]
HARTREE_TO_EV = physical_constants["Hartree energy in eV"][0]
Cperg_to_mAhperg = 2.778e-1
C_TO_mAh = Cperg_to_mAhperg
BOHR_TO_ANGSTROM = physical_constants["Bohr radius"][0] * 1e10
RY_TO_EV = physical_constants["Rydberg constant times hc in eV"][0]
KBAR_TO_GPA = 0.1
eV_PER_ANGSTROM_CUBED_TO_GPa = 160.21776
AVOGADROS_NUMBER = physical_constants["Avogadro constant"][0]
ANGSTROM_CUBED_TO_CENTIMETRE_CUBED = 1e-24
ELECTRON_CHARGE = physical_constants["elementary charge"][0]
KELVIN_TO_EV = physical_constants["kelvin-electron volt relationship"][0]
INVERSE_CM_TO_EV = (
    physical_constants["inverse meter-electron volt relationship"][0] * 100
)
EFG_AU_TO_SI = physical_constants["atomic unit of electric field gradient"][0]
BARN_TO_M2 = 1e-28
PLANCK_CONSTANT = physical_constants["Planck constant"][0]


__all__ = [
    "FARADAY_CONSTANT_Cpermol",
    "HARTREE_TO_EV",
    "Cperg_to_mAhperg",
    "C_TO_mAh",
    "BOHR_TO_ANGSTROM",
    "RY_TO_EV",
    "KBAR_TO_GPA",
    "eV_PER_ANGSTROM_CUBED_TO_GPa",
    "AVOGADROS_NUMBER",
    "ANGSTROM_CUBED_TO_CENTIMETRE_CUBED",
    "ELECTRON_CHARGE",
    "KELVIN_TO_EV",
    "INVERSE_CM_TO_EV",
]



# coding: utf-8
# Distributed under the terms of the MIT license.

import unittest
import numpy as np
import pytest


class TestMagresReferencer(unittest.TestCase):
    def test_fit_only(self):
        from matador.magres import MagresReferencer
        from matador.crystal import Crystal

        SiO2 = Crystal(
            dict(
                lattice_cart=[[10, 0, 0], [0, 10, 0], [0, 0, 10]],
                positions_frac=[[0.0, 0.0, 0.0], [0.1, 0.1, 0.1], [0.2, 0.2, 0.2]],
                atom_types=["Si", "O", "O"],
                chemical_shielding_isos=[-100, -200, -300],
            )
        )

        expt_shifts = {"SiO2": {"O": [100, 200]}}

        ref = MagresReferencer({"SiO2": SiO2}, expt_shifts, species="O")

        self.assertAlmostEqual(ref.fit_gradient, -1)
        self.assertAlmostEqual(ref.fit_intercept, -100)
        self.assertAlmostEqual(ref.fit_rsquared, 1)

    def test_assign_shifts(self):
        from matador.magres import MagresReferencer
        from matador.crystal import Crystal

        SiO2 = Crystal(
            dict(
                lattice_cart=[[10, 0, 0], [0, 10, 0], [0, 0, 10]],
                positions_frac=[[0.0, 0.0, 0.0], [0.1, 0.1, 0.1], [0.2, 0.2, 0.2]],
                atom_types=["Si", "O", "O"],
                chemical_shielding_isos=[-100, -200, -300],
            )
        )

        LiCoO3_supercell = Crystal(
            dict(
                lattice_cart=[[10, 0, 0], [0, 10, 0], [0, 0, 10]],
                positions_frac=[
                    [0.0, 0.0, 0.0],
                    [0.1, 0.1, 0.1],
                    [0.2, 0.2, 0.2],
                    [0.3, 0.3, 0.3],
                    [0.4, 0.4, 0.4],
                    [0.5, 0.5, 0.5],
                    [0.6, 0.6, 0.6],
                    [0.7, 0.7, 0.7],
                    [0.8, 0.8, 0.8],
                    [0.9, 0.9, 0.9],
                ],
                atom_types=["Li", "Li", "Co", "Co", "O", "O", "O", "O", "O", "O"],
                chemical_shielding_isos=[
                    -100,
                    -120,
                    -321,
                    -992,
                    -1500,
                    -1600,
                    -1700,
                    -1800,
                    -1900,
                    -2000,
                ],
            )
        )

        LiCoO3_theory = Crystal(
            dict(
                lattice_cart=[[10, 0, 0], [0, 10, 0], [0, 0, 10]],
                positions_frac=[
                    [0.0, 0.0, 0.0],
                    [0.1, 0.1, 0.1],
                    [0.2, 0.2, 0.2],
                    [0.3, 0.3, 0.3],
                    [0.4, 0.4, 0.4],
                    [0.5, 0.5, 0.5],
                    [0.6, 0.6, 0.6],
                    [0.7, 0.7, 0.7],
                    [0.8, 0.8, 0.8],
                    [0.9, 0.9, 0.9],
                ],
                atom_types=["Li", "Li", "Co", "Co", "O", "O", "O", "O", "O", "O"],
                chemical_shielding_isos=[-100, -120, -321, -992, 1, 2, 3, 4, 5, 6],
            )
        )

        expt_shifts = {
            "SiO2": {"O": [100, 200]},
            "LiCoO3": {"O": [1400, 1500, 1600, 1700, 1800, 1900]},
        }

        ref = MagresReferencer(
            {"SiO2": SiO2, "LiCoO3": LiCoO3_supercell},
            expt_shifts,
            species="O",
            structures=[LiCoO3_theory],
        )

        self.assertAlmostEqual(ref.fit_gradient, -1)
        self.assertAlmostEqual(ref.fit_intercept, -100)
        self.assertAlmostEqual(ref.fit_rsquared, 1)

        LiCoO3_theory = ref.structures[0]
        shifts = [
            site["chemical_shift_iso"] for site in LiCoO3_theory if site.species == "O"
        ]

        np.testing.assert_array_almost_equal(
            shifts, [-101, -102, -103, -104, -105, -106]
        )

        np.testing.assert_array_almost_equal(
            LiCoO3_theory["chemical_shift_isos"][4:],
            [-101, -102, -103, -104, -105, -106],
        )

        with pytest.raises(AttributeError):
            LiCoO3_theory[-1]["chemical_shift_iso"] = 123

        with pytest.raises(ValueError):
            LiCoO3_theory[0][-1]["chemical_shift_iso"] = 123



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the Crystal class, a wrapper
to the raw dictionary stored in MongoDB that allows for validation,
manipulation and analysis of the lattice.

"""

from __future__ import annotations

from copy import deepcopy
from typing import List, Tuple, Union
from matador.utils import cell_utils
from matador.orm.orm import DataContainer
from matador.crystal.crystal_site import Site
from matador.utils.chem_utils import get_concentration, get_subscripted_formula
from matador.utils.cell_utils import real2recip


class UnitCell:
    """This class describes a 3D periodic unit cell by its
    Cartesian lattice vectors or lattice parameters, in Å.

    """

    _lattice_abc = None
    _lattice_cart = None
    _volume = None

    def __init__(self, lattice):
        """Initialise the cell from either Cartesian lattice vectors
        [[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]], or lattice
        parameters [[a, b, c], [alpha, beta, gamma]].

        Parameters:
            lattice (:obj:`list` or numpy.ndarray): either Cartesian
                lattice vectors or lattice parameters, stored as lists
                or a numpy arrays.

        """

        self._volume = None
        if len(lattice) == 3:
            if any(len(vec) != 3 for vec in lattice):
                raise RuntimeError(
                    "Unable to cast {} into lattice_cart".format(lattice)
                )
            self.lattice_cart = lattice
        elif len(lattice) == 2:
            if any(len(vec) != 3 for vec in lattice):
                raise RuntimeError("Unable to cast {} into lattice_abc".format(lattice))
            self.lattice_abc = lattice
        else:
            raise RuntimeError(
                "Unable to create UnitCell from lattice {}".format(lattice)
            )

    @property
    def lattice_cart(
        self,
    ) -> Tuple[
        Tuple[float, float, float],
        Tuple[float, float, float],
        Tuple[float, float, float],
    ]:
        """The Cartesian lattice vectors as a tuple."""
        return self._lattice_cart

    @lattice_cart.setter
    def lattice_cart(self, new_lattice):
        self._lattice_cart = tuple(tuple(vec) for vec in new_lattice)
        self._lattice_abc = tuple(
            tuple(elem) for elem in cell_utils.cart2abc(self._lattice_cart)
        )
        self._volume = None

    @property
    def lattice_abc(
        self,
    ) -> Tuple[Tuple[float, float, float], Tuple[float, float, float]]:
        """Lattice parameters as a tuple."""
        return self._lattice_abc

    @lattice_abc.setter
    def lattice_abc(self, new_lattice):
        self._lattice_abc = tuple(tuple(vec) for vec in new_lattice)
        self._lattice_cart = tuple(
            tuple(vec) for vec in cell_utils.abc2cart(self._lattice_abc)
        )
        self._volume = None

    @property
    def lengths(self) -> Tuple[float, float, float]:
        """Lattice vector lengths."""
        return self._lattice_abc[0]

    @lengths.setter
    def lengths(self, new_lengths):
        if len(new_lengths) != 3:
            raise RuntimeError(
                "Expected list of 3 floats for cell vector lengths, received {}".format(
                    new_lengths
                )
            )
        self.lattice_abc = [new_lengths, self._lattice_abc[1]]

    @property
    def recip_lattice_cart(self) -> List[List[float]]:
        return real2recip(self.lattice_cart)

    @property
    def angles(self) -> Tuple[float, float, float]:
        """Lattice vector angles."""
        return self._lattice_abc[1]

    @angles.setter
    def angles(self, new_angles):
        if len(new_angles) != 3:
            raise RuntimeError(
                "Expected list of 3 floats for cell vector angles, received {}".format(
                    new_angles
                )
            )
        self.lattice_abc = [self._lattice_abc[0], new_angles]

    @property
    def volume(self) -> float:
        """The cell volume in Å³."""
        if not self._volume:
            self._volume = cell_utils.cart2volume(self._lattice_cart)
        return self._volume

    def __str__(self) -> str:
        return (
            f"(a, b, c) = {self.lengths[0]:4.4f} Å, {self.lengths[1]:4.4f} Å, {self.lengths[2]:4.4f} Å\n"
            f"(α, β, γ) = {self.angles[0]:4.4f}° {self.angles[1]:4.4f}° {self.angles[2]:4.4f}°\n"
        )


class Crystal(DataContainer):
    """Class that wraps the MongoDB document, providing useful
    interfaces for cell manipulation and validation.

    Attributes:
        elems (:obj:`list` of :obj:`str`): list of present elements in the crystal,
            sorted in alphabetical order.
        cell (:obj:`UnitCell`): the unit cell constructed from lattice vectors.

    """

    @staticmethod
    def _validate_doc(doc):
        if not any(key in doc for key in ["lattice_cart", "lattice_abc"]):
            raise RuntimeError("No lattice information found, cannot create Crystal.")
        if "atom_types" not in doc:
            raise RuntimeError(
                'No species information found `"atom_types"`, cannot create Crystal.'
            )
        if not any(key in doc for key in ["positions_frac", "positions_abs"]):
            raise RuntimeError(
                'No position information found `"positions_frac"/"positions_abs"`, cannot create Crystal.'
            )
        if "positions_frac" in doc and len(doc["atom_types"]) != len(
            doc["positions_frac"]
        ):
            raise RuntimeError(
                "Position data and atom species data are incompatible: "
                f'{len(doc["positions_frac"])} vs {len(doc["atom_types"])}'
            )
        if "positions_abs" in doc and len(doc["atom_types"]) != len(
            doc["positions_abs"]
        ):
            raise RuntimeError(
                "Position data and atom species data are incompatible: "
                f'{len(doc["positions_abs"])} vs {len(doc["atom_types"])}'
            )

    def __init__(self, doc, voronoi=False, network_kwargs=None, mutable=False):
        """Initialise Crystal object from matador document with Site list
        and any additional abstractions, e.g. voronoi or CrystalGraph.

        Parameters:
            doc (dict): document containing structural information, minimal requirement
                is for `atom_types`, one of `lattice_abc` or `lattice_cart`, and one of
                `positions_frac` or `positions_abs` to be present.

        Keyword Arguments:
           voronoi (bool): whether to compute Voronoi substructure for each site
           network_kwargs (dict): keywords to pass to the CrystalGraph initialiser

        """

        self._validate_doc(doc)
        if isinstance(doc, Crystal):
            doc = deepcopy(doc._data)

        super().__init__(doc, mutable=mutable)

        self.elems = sorted(list(set(self._data["atom_types"])))
        self.sites = []

        # use lattice_cart to construct cell if present, otherwise abc
        self.cell = UnitCell(doc.get("lattice_cart", doc.get("lattice_abc")))

        self._construct_sites(voronoi=voronoi)

        if network_kwargs is not None:
            self._network_kwargs = network_kwargs
        else:
            self._network_kwargs = {}

        # assign attributes for later
        self._coordination_lists = None
        self._coordination_stats = None
        self._network = None
        self._bond_lengths = None
        self._bonding_stats = None

        # assume default value for symprec
        if "space_group" in self._data:
            self._space_group = {0.01: self._data["space_group"]}
        else:
            self._space_group = {}

    def __getitem__(self, key: Union[str, int]):
        """If integer key is requested, return index into site array. If
        a known site data key is requested, return the live version attached
        to the sites. Otherwise, go through the DataContainer methods.

        """
        if isinstance(key, int):
            return self.sites[key]

        elif key in Site._crystal_key_map:
            return [s.get(Site._crystal_key_map[key]) for s in self]

        return super().__getitem__(key)

    def __str__(self) -> str:
        repr_string = f"{self.formula_unicode}: {self.root_source}\n"
        repr_string += (len(repr_string) - 1) * "=" + "\n"
        repr_string += f"{self.num_atoms:<3} atoms. {self.space_group}\n"

        if "formation_enthalpy_per_atom" in self._data:
            repr_string += "Formation enthalpy = {:6.6f} eV/atom\n".format(
                self._data["formation_enthalpy_per_atom"]
            )

        repr_string += self.cell.__str__()

        return repr_string

    def __repr__(self) -> str:
        return f"<Crystal: {self.formula_unicode} {self.root_source}>"

    def update(self, data):
        """Update the underlying `self._data` dictionary
        with the passed data.

        """
        self._data.update(data)

    def print_sites(self):
        print("---")
        for ind, site in enumerate(self):
            print(f"{ind:3d}", end=" ")
            print(site)

    def set_positions(self, new_positions, fractional=True):
        if len(new_positions) != self.num_atoms:
            raise RuntimeError("Cannot change size of positions array!")
        if fractional:
            self._data["positions_frac"] = new_positions
            self._data.pop("positions_abs", None)
        else:
            self._data["positions_abs"] = new_positions
            self._data.pop("positions_frac", None)

        self._construct_sites()

    def _construct_sites(self, voronoi=False):
        """Constructs the list of Site objects stored in self.sites.

        Keyword arguments:
            voronoi (bool): whether to calculate the Voronoi substructure
                of each site.

        """

        self.sites = []
        for ind, species in enumerate(self.atom_types):
            position = self.positions_frac[ind]
            site_data = {}
            for key in Site._crystal_key_map:
                if key in self._data and len(self._data[key]) == len(
                    self._data["atom_types"]
                ):
                    site_data[Site._crystal_key_map[key]] = self._data[key][ind]

            if voronoi and "voronoi_substructure" not in site_data:
                site_data["voronoi_substructure"] = self.voronoi_substructure[ind]

            self.sites.append(Site(species, position, self.cell, **site_data))

    @property
    def atom_types(self) -> List[str]:
        """Return list of atom types."""
        return self._data["atom_types"]

    @property
    def num_atoms(self) -> int:
        """Return number of atoms in structure."""
        return len(self.sites)

    @property
    def num_elements(self) -> int:
        """Return number of species in the structure."""
        return len(self.elems)

    @property
    def positions_frac(self) -> List[List[float]]:
        """Return list of fractional positions."""
        from matador.utils.cell_utils import cart2frac

        if "positions_frac" not in self._data:
            self._data["positions_frac"] = cart2frac(
                self.cell.lattice_cart, self.positions_abs
            )
        return self._data["positions_frac"]

    @property
    def positions_abs(self) -> List[List[float]]:
        """Return list of absolute Cartesian positions."""
        from matador.utils.cell_utils import frac2cart

        if "positions_abs" not in self._data:
            self._data["positions_abs"] = frac2cart(
                self.cell.lattice_cart, self.positions_frac
            )
        return self._data["positions_abs"]

    @property
    def site_occupancies(self) -> List[float]:
        """Return a list of site occupancies."""
        if "site_occupancies" not in self._data:
            self._data["site_occupancies"] = [site.occupancy for site in self]
        return self._data["site_occupancies"]

    @property
    def stoichiometry(self) -> List[List[Union[str, float]]]:
        """Return stoichiometry in matador format: a list of
        two-member lists containing element symbol and number
        of atoms per formula unit, sorted in alphabetical order
        by element symbol).

        """
        if "stoichiometry" not in self._data:
            from matador.utils.chem_utils import get_stoich

            self._data["stoichiometry"] = get_stoich(self.atom_types)
        return self._data["stoichiometry"]

    @property
    def num_fu(self) -> int:
        if "num_fu" not in self._data:
            self._data["num_fu"] = self.num_atoms / sum(
                atom[1] for atom in self.stoichiometry
            )
        return self._data["num_fu"]

    @property
    def concentration(self) -> List[float]:
        """Return concentration of each species in stoichiometry."""
        if "concentration" not in self._data:
            self._data["concentration"] = get_concentration(
                self.stoichiometry,
                [elem[0] for elem in self.stoichiometry],
                include_end=True,
            )
        return self._data["concentration"]

    def get_concentration(self, elements=None, include_end=True) -> List[float]:
        """Return concentration of each species in stoichiometry, and use this
        value for future calls to `crystal.concentration`.

        """
        if elements is None:
            elements = self.elems
        self._data["concentration"] = get_concentration(
            self.stoichiometry, elements=elements, include_end=include_end
        )
        return self._data["concentration"]

    @property
    def formula(self) -> str:
        """Returns chemical formula of structure."""
        from matador.utils.chem_utils import get_formula_from_stoich

        return get_formula_from_stoich(self.stoichiometry, tex=False)

    @property
    def formula_tex(self) -> str:
        """Returns chemical formula of structure in LaTeX format."""
        from matador.utils.chem_utils import get_formula_from_stoich

        return get_formula_from_stoich(self.stoichiometry, tex=True)

    @property
    def formula_unicode(self) -> str:
        """Returns a unicode string for the chemical formula."""
        return get_subscripted_formula(self.formula)

    @property
    def cell_volume(self) -> float:
        """Returns cell volume in Å³."""
        return self.cell.volume

    @property
    def lattice_cart(
        self,
    ) -> Tuple[
        Tuple[float, float, float],
        Tuple[float, float, float],
        Tuple[float, float, float],
    ]:
        """The Cartesian lattice vectors as a tuple."""
        return self.cell.lattice_cart

    @property
    def lattice_abc(
        self,
    ) -> Tuple[Tuple[float, float, float], Tuple[float, float, float]]:
        """Lattice parameters as a tuple."""
        return self.cell.lattice_abc

    @property
    def bond_lengths(self) -> List[Tuple[Tuple[str, str], float]]:
        """Returns a list of ((species_A, species_B), bond_length)),
        sorted by bond length, computed from the network structure of
        the crystal (i.e. first coordination sphere).

        """
        if self._bond_lengths is None:
            self._bond_lengths = []
            for i, j, data in self.network.edges.data():
                self._bond_lengths.append(
                    ((self[i].species, self[j].species), data["dist"])
                )
            self._bond_lengths = sorted(self._bond_lengths, key=lambda bond: bond[1])
        return self._bond_lengths

    @property
    def voronoi_substructure(self):
        """Returns, or calculates  if not present, the Voronoi
        substructure of crystal.

        """
        if "voronoi_substructure" not in self._data:
            from matador.plugins.voronoi_interface.voronoi_interface import (
                get_voronoi_substructure,
            )

            self._data["voronoi_substructure"] = get_voronoi_substructure(self._data)
        return self._data["voronoi_substructure"]

    @property
    def space_group(self) -> str:
        """Return the space group symbol at the last-used symprec."""
        return self.get_space_group(symprec=self._data.get("symprec", 0.01))

    @property
    def space_group_tex(self) -> str:
        """Returns the space group symbol at the last-used symprec, formatted for LaTeX.

        The HM symbol is surrounded by $ and with '-' replaced with overbars for rotoinversion axes,
        e.g., Fm-3m -> $Fm\\bar{3}m$.
        """
        from matador.utils.cell_utils import get_space_group_label_latex

        return get_space_group_label_latex(self.space_group)

    def get_space_group(self, symprec=0.01) -> str:
        """Return the space group of the structure at the desired
        symprec. Stores the space group in a dictionary
        `self._space_group` under symprec keys. Updates
        `self._data['space_group']` and `self._data['symprec']` with the
        last value calculated.

        Keyword arguments:
            symprec (float): spglib symmetry tolerance.

        """
        if symprec not in self._space_group:
            self._data["space_group"] = cell_utils.get_spacegroup_spg(
                self._data, symprec=symprec, check_occ=False
            )
            self._data["symprec"] = symprec
            self._space_group[symprec] = self._data["space_group"]
        return self._space_group[symprec]

    @property
    def pdf(self):
        """Returns a PDF object (pair distribution function) for the
        structure, calculated with default PDF settings.

        """
        from matador.fingerprints.pdf import PDF

        if "pdf" not in self._data:
            self._data["pdf"] = PDF(
                self._data, label=self.formula_tex, standardize=False
            )
        return self._data["pdf"]

    @pdf.setter
    def pdf(self, pdf):
        """Set the PDF to the given PDF object (or None)."""
        from matador.fingerprints.pdf import PDF

        if isinstance(pdf, PDF) or pdf is None:
            self._data["pdf"] = pdf

    def calculate_pdf(self, **kwargs):
        """Calculate and set the PDF with the passed parameters."""
        from matador.fingerprints.pdf import PDF

        if "pdf" not in self._data:
            self._data["pdf"] = PDF(self._data, label=self.formula_tex, **kwargs)
        return self._data["pdf"]

    @property
    def pxrd(self):
        """Returns a PXRD object (powder xray diffraction) containing
        the XRD pattern for the structure.

        """
        from matador.fingerprints.pxrd import PXRD

        if "pxrd" not in self._data:
            self._data["pxrd"] = PXRD(self._data)
        return self._data["pxrd"]

    @pxrd.setter
    def pxrd(self, pxrd):
        """Set the PXRD to the given PXRD object (or None)."""
        from matador.fingerprints.pxrd import PXRD

        if isinstance(pxrd, PXRD) or pxrd is None:
            self._data["pxrd"] = pxrd

    def calculate_pxrd(self, **kwargs):
        """Compute and set PXRD with the passed parameters."""
        from matador.fingerprints.pxrd import PXRD

        if "pxrd" not in self._data:
            self._data["pxrd"] = PXRD(self._data, **kwargs)
        return self._data["pxrd"]

    @property
    def coordination_stats(self):
        """Returns stastistics on the coordination of each site, as
        computed from Voronoi decomposition.

        """
        if self._coordination_stats is None:
            coordination_lists = self.coordination_lists
            import numpy as np

            coordination_stats = {}
            for species in self.elems:
                coordination_stats[species] = {}
                for _species in self.elems:
                    coordination_stats[species][_species] = {}
                    coordination_stats[species][_species]["mean"] = np.mean(
                        coordination_lists[species][_species]
                    )
                    coordination_stats[species][_species]["median"] = np.median(
                        coordination_lists[species][_species]
                    )
                    coordination_stats[species][_species]["std"] = np.std(
                        coordination_lists[species][_species]
                    )
            self._coordination_stats = coordination_stats

        return self._coordination_stats

    @property
    def ase_atoms(self):
        """Returns an ASE Atoms representation of
        the crystal.

        """
        from matador.utils.ase_utils import doc2ase

        return doc2ase(self, add_keys_to_info=True)

    @property
    def pmg_structure(self):
        """Returns the pymatgen structure representation of
        the crystal.

        """
        from matador.utils.pmg_utils import doc2pmg

        return doc2pmg(self)

    @classmethod
    def from_ase(cls, atoms):
        from matador.utils.ase_utils import ase2dict

        return ase2dict(atoms, as_model=True)

    @classmethod
    def from_pmg(cls, pmg):
        from matador.utils.pmg_utils import pmg2dict

        return pmg2dict(pmg, as_model=True)

    @property
    def coordination_lists(self):
        """Returns a dictionary containing pairs of elements and the
        list of coordination numbers per site.

        """
        if self._coordination_lists is None:
            coordination_lists = {}
            for species in self.elems:
                coordination_lists[species] = {}
                for _species in self.elems:
                    coordination_lists[species][_species] = []
            for site in self:
                for _species in self.elems:
                    if _species in site.coordination:
                        coordination_lists[site.species][_species].append(
                            site.coordination[_species]
                        )
                    else:
                        coordination_lists[site.species][_species].append(0)

            self._coordination_lists = coordination_lists

        return self._coordination_lists

    @property
    def unique_sites(self):
        """Return unique sites using Voronoi decomposition."""
        from matador.plugins.voronoi_interface.voronoi_similarity import (
            get_unique_sites,
        )

        get_unique_sites(self._data)
        return self._data["similar_sites"]

    @property
    def network(self):
        """Returns/constructs a CrystalGraph object of the structure."""
        from matador.crystal.network import CrystalGraph

        if self._network is None:
            self._network = CrystalGraph(self, **self._network_kwargs)
        return self._network

    @property
    def network_stats(self):
        """Return network-calculated coordnation stats."""
        from collections import defaultdict

        coordination = defaultdict(list)
        for node, data in self.network.nodes.data():
            num_edges = len(self.network.edges(node))
            coordination[data["species"]].append(num_edges)
        return coordination

    @property
    def bonding_stats(self):
        """Return network-calculated bonding stats.

        Returns:
            dict: sorted dictionary with root atom as keys and bond
                information as values.

        """
        if self._bonding_stats is None:
            from collections import defaultdict

            bonding_dict = defaultdict(dict)
            network = self.network
            for node in network.nodes:
                bonding_dict[node] = {
                    "species": self.sites[node].species,
                    "position": self.sites[node].coords,
                    "bonds": [],
                }
            bonds = set()
            for data in network.edges.data():
                atom_1 = data[0]
                atom_2 = data[1]
                pair = tuple(sorted([atom_1, atom_2]))
                if pair in bonds:
                    continue
                else:
                    bonds.add(pair)

                site_1 = self.sites[atom_1]
                site_2 = self.sites[atom_2]

                bond_length = data[2]["dist"]
                is_image = bool(data[2]["image"])
                bonding_dict[atom_1]["bonds"].append(
                    {
                        "species": site_2.species,
                        "index": atom_2,
                        "length": bond_length,
                        "is_image": is_image,
                        "position": site_2.coords,
                    }
                )
                bonding_dict[atom_2]["bonds"].append(
                    {
                        "species": site_1.species,
                        "index": atom_1,
                        "length": bond_length,
                        "is_image": is_image,
                        "position": site_1.coords,
                    }
                )

            for key in bonding_dict:
                bonding_dict[key]["bonds"] = sorted(
                    bonding_dict[key]["bonds"], key=lambda x: x["index"]
                )

            self._bonding_stats = {
                key: bonding_dict[key] for key in sorted(bonding_dict)
            }

        return self._bonding_stats

    def draw_network(self, layout=None):
        """Draw the CrystalGraph network."""
        from matador.crystal.network import draw_network

        draw_network(self, layout=layout)

    def supercell(self, extension=None, target=None):
        """Returns a supercell of this crystal with the specified extension."""
        if extension is not None:
            from matador.utils.cell_utils import create_simple_supercell

            return create_simple_supercell(self, extension)
        elif target is not None:
            from matador.utils.cell_utils import (
                create_supercell_with_minimum_side_length,
            )

            return create_supercell_with_minimum_side_length(self, target)
        else:
            raise RuntimeError("One of `extension` or `target` must be specified.")

    def standardized(self, primitive=False):
        """Returns a standardized cell for this crystal, optionally the primitive cell."""
        from matador.utils.cell_utils import standardize_doc_cell

        return standardize_doc_cell(self, primitive=primitive)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements plotting routines specifically
for the PXRD objects defined in the
matador.fingerprints.pxrd module.

"""


from matador.plotting.plotting import plotting_function
from matador.utils.cell_utils import get_space_group_label_latex
from matador.crystal import Crystal


__all__ = ["plot_pxrd"]


@plotting_function
def plot_pxrd(
    pxrds,
    two_theta_range=None,
    text_loc="right",
    rug=False,
    rug_height=0.05,
    rug_offset=0.04,
    offset=None,
    ax=None,
    labels=None,
    figsize=None,
    text_offset=0.1,
    colour_labels=True,
    filename=None,
    **kwargs,
):
    """Plot PXRD or PXRDs.

    Parameters:
        pxrds (list or matador.fingerprints.pxrd.PXRD): the PXRD
            or list of PXRDs to plot.

    Keyword arguments:
        two_theta_range (tuple): plotting limits for 2theta
        rug (bool): whether to provide a rug plot of all peaks.
        rug_height (float): size of rug ticks.
        rug_offset (float): offset of rug ticks.
        offset (float): extra space added between patterns (as fraction
            of max intensity). Default 0.1.
        labels (list of str): list of labels to plot alongside pattern.
        figsize (tuple): specify a figure size, the default
            scales with the number of PXRDs to be plotted.
        ax (matplotlib.Axis): optional axis object to plot on.
        text_offset (float): amount by which to offset the labels.
        text_loc (str): location of text, either 'left' or 'right'.
        filename (str): optional filename for saving.


    """

    import matplotlib.pyplot as plt

    if not isinstance(pxrds, list):
        pxrds = [pxrds]
    if labels is not None and not isinstance(labels, list):
        labels = [labels]

    if figsize is None and ax is None:
        _user_default_figsize = plt.rcParams.get("figure.figsize", (8, 6))
        height = len(pxrds) * max(0.5, _user_default_figsize[1] / 1.5 / len(pxrds))
        figsize = (_user_default_figsize[0], height)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if offset is None and rug:
        offset = 0.2
    elif offset is None:
        offset = 0.1

    colour_cycle = ax._get_lines.prop_cycler

    for ind, pxrd in enumerate(pxrds):
        if isinstance(pxrd, Crystal):
            pxrd = pxrd.pxrd
        elif isinstance(pxrd, dict) and "pxrd" in pxrd:
            pxrd = pxrd["pxrd"]

        c = next(colour_cycle).get("color")

        if labels:
            label = labels[ind]
        else:
            label = f"{pxrd.formula}-{get_space_group_label_latex(pxrd.spg)}"

        ax.plot(pxrd.two_thetas, (1 - offset) * pxrd.pattern + ind, c=c)

        if colour_labels:
            text_colour = c
        else:
            text_colour = None

        if text_loc == "right":
            ax.text(
                0.95,
                ind + text_offset,
                label,
                transform=ax.get_yaxis_transform(),
                horizontalalignment="right",
                color=text_colour,
            )

        if text_loc == "left":
            ax.text(
                0.05,
                ind + text_offset,
                label,
                transform=ax.get_yaxis_transform(),
                horizontalalignment="left",
                color=text_colour,
            )

        if rug:
            import numpy as np

            peaks = np.unique(pxrd.peak_positions)
            for peak in peaks:
                ax.plot(
                    [peak, peak],
                    [ind - rug_height - rug_offset, ind - rug_offset],
                    c=c,
                    alpha=0.5,
                )

    ax.set_ylim(-0.2, len(pxrds) + 0.1)
    if two_theta_range is not None:
        ax.set_xlim(*two_theta_range)
    ax.set_ylabel("Relative intensity")
    ax.set_xlabel("$2\\theta$ (degrees)")

    return ax



# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements the BatchRun class for chaining
ComputeTask instances across several structures with
high-throughput.

"""


from collections import defaultdict
import multiprocessing as mp
import os
import glob
import time
import random
import psutil
from matador.utils.print_utils import print_failure, print_warning
from matador.compute.queueing import get_queue_manager
from matador.scrapers.castep_scrapers import cell2dict, param2dict
from matador.compute.compute import ComputeTask
from matador.utils.errors import (
    InputError,
    CalculationError,
    MaxMemoryEstimateExceeded,
    NodeCollisionError,
)


class BatchRun:
    """A class that implements the running of multiple generic jobs on
    a series of files without collisions with other nodes using the
    ComputeTask class. Jobs that have been started are listed in
    ``jobs.txt``, failed jobs are moved to ``bad_castep/``, completed jobs
    are moved to ``completed/``.

    Interface initially inspired by on run.pl, run2.pl and PyAIRSS class
    CastepRunner.

    """

    def __init__(self, seed, **kwargs):
        """Check directory has valid contents and prepare log files
        and directories if not already prepared, then begin running
        calculations.

        Note:
            This class is usually initialised by the run3 script, which
            has a full description of possible arguments.

        Parameters:
            seed (:obj:`list` of :obj:`str`): single entry of param/cell
                file seed for CASTEP geometry optimisations of res
                files, or a list of filenames of ``$seed`` to run arbitrary
                executables on. e.g. ``['LiAs']`` if LiAs.cell and LiAs.param
                exist in cwd full of res files, e.g.2. ``['LiAs_1', 'LiAs_2']``
                if LiAs_1.in/LiAs_2.in exist, and executable = 'pw6.x < $seed.in'.

        Keyword arguments:
            Exhaustive list found in argparse parser inside `matador/cli/run3.py`.

        """
        # parse args, then co-opt them for passing directly into ComputeTask
        prop_defaults = {
            "ncores": None,
            "nprocesses": 1,
            "nnodes": 1,
            "executable": "castep",
            "no_reopt": False,
            "mode": None,
            "redirect": None,
            "debug": False,
            "custom_params": False,
            "verbosity": 0,
            "archer": False,
            "slurm": False,
            "intel": False,
            "conv_cutoff": False,
            "conv_kpt": False,
            "memcheck": False,
            "maxmem": None,
            "killcheck": True,
            "scratch_prefix": None,
            "kpts_1D": False,
            "spin": None,
            "ignore_jobs_file": False,
            "rough": 4,
            "rough_iter": 2,
            "fine_iter": 20,
            "max_walltime": None,
            "limit": None,
            "profile": False,
            "polltime": 30,
        }
        self.args = {}
        self.args.update(prop_defaults)
        self.args.update(kwargs)
        self.debug = self.args.get("debug")

        self.seed = seed
        # if only one seed, check if it is a file, and if so treat
        # this run as a generic run, not a CASTEP cell/param run
        if len(self.seed) == 1 and isinstance(self.seed, list):
            if "*" in self.seed[0]:
                self.seed = glob.glob(self.seed[0])
            elif not os.path.isfile(self.seed[0]):
                self.seed = self.seed[0]

        self.compute_dir = os.uname()[1]
        if self.args.get("scratch_prefix") not in [None, "."]:
            self.compute_dir = "{}/{}".format(
                self.args["scratch_prefix"], self.compute_dir
            ).replace("//", "/")
        elif self.args.get("scratch_prefix") == ".":
            self.compute_dir = None

        if self.args.get("mode") is not None:
            self.mode = self.args.get("mode")
        else:
            if isinstance(self.seed, str):
                self.mode = "castep"
            else:
                self.mode = "generic"
        del self.args["mode"]

        if self.args.get("no_reopt"):
            self.args["reopt"] = False
        else:
            self.args["reopt"] = True
        if "no_reopt" in self.args:
            del self.args["no_reopt"]
        self.nprocesses = int(self.args["nprocesses"])
        del self.args["nprocesses"]
        self.limit = self.args.get("limit")
        del self.args["limit"]
        self.maxmem = self.args.get("maxmem")
        del self.args["maxmem"]
        self.max_walltime = self.args.get("max_walltime")

        # detect and scrape queue settings
        self.queue_mgr = get_queue_manager()

        if self.queue_mgr is not None:
            if self.maxmem is None:
                self.maxmem = self.queue_mgr.max_memory
            if self.max_walltime is None:
                self.max_walltime = self.queue_mgr.walltime

        self.start_time = None
        if self.max_walltime is not None:
            self.start_time = time.time()

        # assign number of cores
        self.all_cores = psutil.cpu_count(logical=False)
        if self.args.get("ncores") is None:
            if self.queue_mgr is None:
                self.args["ncores"] = int(self.all_cores / self.nprocesses)
            else:
                self.args["ncores"] = int(self.queue_mgr.ntasks / self.nprocesses)

        if self.args["nnodes"] < 1 or self.args["ncores"] < 1 or self.nprocesses < 1:
            raise InputError("Invalid number of cores, nodes or processes.")
        if self.all_cores < self.nprocesses:
            raise InputError(
                "Requesting more processes than available cores: {} vs {}".format(
                    self.all_cores, self.nprocesses
                )
            )

        # scrape input cell/param/other files
        if self.mode == "castep":
            self.castep_setup()
        else:
            self.generic_setup()

        # prepare folders and text files
        self.paths = dict()
        if self.args.get("conv_cutoff"):
            self.paths["completed_dir"] = "completed_cutoff"
        elif self.args.get("conv_kpt"):
            self.paths["completed_dir"] = "completed_kpts"
        else:
            self.paths["completed_dir"] = "completed"
        self.paths["failed_dir"] = "bad_castep"
        self.paths["jobs_fname"] = "jobs.txt"
        self.paths["completed_fname"] = "finished_cleanly.txt"
        self.paths["failures_fname"] = "failures.txt"
        self.paths["memory_fname"] = "memory_exceeded.txt"
        if not os.path.isfile(self.paths["jobs_fname"]):
            with open(self.paths["jobs_fname"], "a"):
                pass
        if not os.path.isfile(self.paths["completed_fname"]):
            with open(self.paths["completed_fname"], "a"):
                pass
        if not os.path.isfile(self.paths["failures_fname"]):
            with open(self.paths["failures_fname"], "a"):
                pass
        if self.args.get("memcheck"):
            if not os.path.isfile(self.paths["memory_fname"]):
                with open(self.paths["memory_fname"], "a"):
                    pass

    def spawn(self, join=False):
        """Spawn processes to perform calculations.

        Keyword arguments:
            join (bool): whether or not to attach to ComputeTask
                process. Useful for testing.

        """
        procs = []
        error_queue = mp.Queue()
        for proc_id in range(self.nprocesses):
            procs.append(
                mp.Process(
                    target=self.perform_new_calculations,
                    args=(
                        random.sample(
                            self.file_lists["res"], len(self.file_lists["res"])
                        ),
                        error_queue,
                        proc_id,
                    ),
                )
            )
        for proc in procs:
            proc.start()
            if join:
                proc.join()

        errors = []
        failed_seeds = []

        # wait for each proc to write to error queue
        try:
            for _, proc in enumerate(procs):
                result = error_queue.get()
                if isinstance(result[1], Exception):
                    errors.append(result)
                    failed_seeds.append(result[2])

            if errors:
                error_message = ""
                for error in errors:
                    error_message += "Process {} raised error(s): {}. ".format(
                        error[0], error[1]
                    )
                    if len({type(error[1]) for error in errors}) == 1:
                        raise errors[0][1]
                    raise type(errors[0][1])(error_message)
                raise BundledErrors(error_message)
        # the only errors that reach here are fatal, e.g. WalltimeError, CriticalError, InputError, KeyboardInterrupt
        except RuntimeError as err:
            result = [proc.join(timeout=2) for proc in procs]
            result = [proc.terminate() for proc in procs if proc.is_alive()]
            try:
                # Guard against failures to write to stdout
                print_failure("Fatal error(s) reported:")
                print_warning(err)
            except OSError:
                pass
            raise err

        try:
            # Guard against failures to write to stdout
            print("Nothing left to do.")
        except OSError:
            pass

    def perform_new_calculations(self, res_list, error_queue, proc_id):
        """Perform all calculations that have not already
        failed or finished to completion.

        Parameters:
            res_list (:obj:`list` of :obj:`str`): list of structure filenames.
            error_queue (multiprocessing.Queue): queue to push exceptions to
            proc_id (int): process id for logging

        """
        job_count = 0
        if isinstance(res_list, str):
            res_list = [res_list]
        for res in res_list:
            try:
                if not os.path.isfile(res):
                    continue
                # probe once then sleep for a random amount up to 5 seconds
                # before checking again for a lock file, just to protect
                # against collisions in large array jobs on slower parallel file systems
                _ = os.path.isfile("{}.lock".format(res))
                time.sleep(2 * random.random())
                # wait some additional time if this is a slurm array job
                if self.queue_mgr is not None:
                    extra_wait = (
                        self.queue_mgr.array_id % 10 if self.queue_mgr.array_id else 0
                    )
                    time.sleep(extra_wait)
                locked = os.path.isfile("{}.lock".format(res))
                if not self.args.get("ignore_jobs_file"):
                    listed = self._check_jobs_file(res)
                else:
                    listed = []
                running = any([listed, locked])
                if not running:
                    # check we haven't reached job limit
                    if self.limit is not None and job_count >= self.limit:
                        error_queue.put((proc_id, job_count, res))
                        return

                    # check 3 more times if a lock exists with random up to 1 second
                    # waits each time
                    for _ in range(3):
                        time.sleep(random.random())
                        if os.path.isfile("{}.lock".format(res)):
                            raise NodeCollisionError(
                                "Another node wrote this file when I wanted to, skipping..."
                            )
                    with open(res + ".lock", "a") as job_file:
                        pass

                    # write to jobs file
                    with open(self.paths["jobs_fname"], "a") as job_file:
                        job_file.write(res + "\n")

                    # create full relaxer object for creation and running of job
                    job_count += 1
                    relaxer = ComputeTask(
                        node=None,
                        res=res,
                        param_dict=self.param_dict,
                        cell_dict=self.cell_dict,
                        mode=self.mode,
                        paths=self.paths,
                        compute_dir=self.compute_dir,
                        timings=(self.max_walltime, self.start_time),
                        maxmem=self.maxmem,
                        **self.args,
                    )
                    # if memory check failed, let other nodes have a go
                    if not relaxer.enough_memory:
                        with open(self.paths["memory_fname"], "a") as job_file:
                            job_file.write(res + "\n")
                        if os.path.isfile("{}.lock".format(res)):
                            os.remove("{}.lock".format(res))
                        with open(self.paths["jobs_fname"], "r+") as job_file:
                            flines = job_file.readlines()
                            job_file.seek(0)
                        for line in flines:
                            if res not in line:
                                job_file.write(line)
                            job_file.truncate()

                    elif relaxer.success:
                        with open(self.paths["completed_fname"], "a") as job_file:
                            job_file.write(res + "\n")
                    else:
                        with open(self.paths["failures_fname"], "a") as job_file:
                            job_file.write(res + "\n")

            # catch memory errors and reset so another node can try
            except MaxMemoryEstimateExceeded:
                reset_single_seed(res)
                continue

            # ignore any other individual calculation errors or node collisions that were caught here
            except CalculationError:
                continue

            # reset txt/lock for an input error, but throw it to prevent other calcs
            except InputError as err:
                reset_single_seed(res)
                error_queue.put((proc_id, err, res))
                return
            # push globally-fatal errors to queue, and return to prevent further calcs
            except RuntimeError as err:
                error_queue.put((proc_id, err, res))
                return
            # finally catch any other generic error in e.g. the above code, normally caused by me
            except Exception as err:
                error_queue.put((proc_id, err, res))
                return

        error_queue.put((proc_id, job_count, ""))

    def generic_setup(self):
        """Undo things that are set ready for CASTEP jobs..."""
        self.cell_dict = None
        self.param_dict = None

        # scan directory for files to run
        self.file_lists = defaultdict(list)
        self.file_lists["res"] = glob.glob("*.res")

    def castep_setup(self):
        """Set up CASTEP jobs from res files, and $seed.cell/param."""
        # read cell/param files
        exts = ["cell", "param"]
        for ext in exts:
            if not os.path.isfile("{}.{}".format(self.seed, ext)):
                raise InputError(
                    "Failed to find {ext} file, {seed}.{ext}".format(
                        ext=ext, seed=self.seed
                    )
                )
        self.cell_dict, cell_success = cell2dict(
            self.seed + ".cell", db=False, lattice=False, positions=True
        )
        if not cell_success:
            raise InputError(f"Failed to parse cell file: {self.cell_dict}")
        self.param_dict, param_success = param2dict(self.seed + ".param", db=False)
        if not param_success:
            raise InputError(f"Failed to parse param file: {self.param_dict}")

        # scan directory for files to run
        self.file_lists = defaultdict(list)
        self.file_lists["res"] = glob.glob("*.res")
        if any(
            self.seed == file.replace(".res", "") for file in self.file_lists["res"]
        ):
            error = "Found .res file with same name as seed: {}.res. This will wreak havoc on your calculations!\n".format(
                self.seed
            ) + "Please rename either your seed.cell/seed.param files, or rename the offending {}.res".format(
                self.seed
            )
            raise InputError(error)

        if not self.file_lists["res"]:
            error = "run3 in CASTEP mode requires at least 1 res file in folder, found {}".format(
                len(self.file_lists["res"])
            )
            raise InputError(error)

        if len(self.file_lists["res"]) < self.nprocesses and not any(
            [self.args.get("conv_cutoff"), self.args.get("conv_kpt")]
        ):
            raise InputError("Requested more processes than there are jobs to run!")

        # do some prelim checks of parameters
        if self.param_dict["task"].upper() in [
            "GEOMETRYOPTIMISATION",
            "GEOMETRYOPTIMIZATION",
        ]:
            if "geom_max_iter" not in self.param_dict:
                raise InputError("geom_max_iter is unset, please fix this.")
            if int(self.param_dict["geom_max_iter"]) <= 0:
                raise InputError(
                    "geom_max_iter is only {}!".format(self.param_dict["geom_max_iter"])
                )

        # parse convergence args and set them up
        self.convergence_run_setup()

        # delete source from cell and param
        del self.cell_dict["source"]
        del self.param_dict["source"]

    def convergence_run_setup(self):
        """Set the correct args for a convergence run."""
        # check if we're doing a conv run
        if self.args.get("conv_cutoff"):
            if os.path.isfile("cutoff.conv"):
                with open("cutoff.conv", "r") as f:
                    flines = f.readlines()
                    self.args["conv_cutoff"] = []
                    for line in flines:
                        if not line.startswith("#"):
                            self.args["conv_cutoff"].append(int(line))
            else:
                raise InputError("Missing cutoff.conv file")
        else:
            self.args["conv_cutoff"] = None

        if self.args.get("conv_kpt"):
            if os.path.isfile("kpt.conv"):
                with open("kpt.conv", "r") as f:
                    flines = f.readlines()
                    self.args["conv_kpt"] = []
                    for line in flines:
                        if not line.startswith("#"):
                            self.args["conv_kpt"].append(float(line))
            else:
                raise InputError("Missing with conv.kpt file")
        else:
            self.args["conv_kpt"] = None

    def _check_jobs_file(self, res):
        """Check if structure is listed in jobs.txt file.

        Parameters:
            res (str): structure name.

        Returns:
            bool: True if already listed in jobs file.

        """
        with open(self.paths["jobs_fname"], "r") as job_file:
            flines = job_file.readlines()
            for line in flines:
                if res in line:
                    return True
        return False


class BundledErrors(Exception):
    """Raise this after collecting all exceptions from
    processes.
    """


def reset_job_folder(debug=False):
    """Remove all lock files and clean up jobs.txt
    ready for job restart.

    Note:
        This should be not called by a ComputeTask instance, in case
        other instances are running.

    Returns:
        num_remaining (int): number of structures left to relax

    """
    res_list = glob.glob("*.res")
    for f in res_list:
        root = f.replace(".res", "")
        exts_to_rm = ["res.lock", "kill"]
        for ext in exts_to_rm:
            if os.path.isfile("{}.{}".format(root, ext)):
                os.remove("{}.{}".format(root, ext))

    # also remove from jobs file
    if os.path.isfile("jobs.txt"):
        with open("jobs.txt", "r+") as f:
            flines = f.readlines()
            f.seek(0)
            for line in flines:
                line = line.strip()
                if line in res_list:
                    continue
                f.write(line)
            f.truncate()
            flines = f.readlines()

    return len(res_list)


def reset_single_seed(seed):
    """Remove the file lock and jobs.txt entry
    for a single seed.

    Parameters:
        seed (str): the seedname to remove.

    """
    if os.path.isfile("jobs.txt"):
        with open("jobs.txt", "r+") as f:
            flines = f.readlines()
            f.seek(0)
            for line in flines:
                line = line.strip()
                if seed in line:
                    continue
                f.write(line)
            f.truncate()
            flines = f.readlines()
    if os.path.isfile(seed + ".lock"):
        os.remove(seed + ".lock")



# coding: utf-8
# Distributed under the terms of the MIT License.

import numpy as np
from matador.utils.cell_utils import real2recip, frac2cart
from matador.orm.orm import DataContainer


class Spectral(DataContainer):
    """
    Note:
        This class speaks of "k-points" as general reciprocal space points
        used to display the dispersion curves; these correspond to CASTEP's
        phonon_kpoints or spectral_kpoints, and not the k-points
        used to generate the underlying wavefunction or dynamical matrix.

    """

    @property
    def eigs(self):
        """Alias for the correct eigenvalue array."""
        if "Vibrational" in self.__class__.__name__:
            return self._data["eigs_q"]
        return self._data["eigs_s_k"]

    @property
    def lattice_cart(self):
        """The Cartesian lattice vectors of the real space lattice."""
        return self._data["lattice_cart"]

    @property
    def num_kpoints(self):
        """Number of dispersion k-points sampled."""
        return self._data["num_kpoints"]

    @property
    def num_qpoints(self):
        """Alias for number of kpoints."""
        return self.num_kpoints

    @property
    def projectors(self):
        """Return list of projector labels in the format
        `(element, l-channel)`.

        """
        return self._data.get("projectors")

    @property
    def num_modes(self):
        """Number of eigenvalues per q/k-point."""
        return self._data["num_modes"]

    @property
    def num_bands(self):
        """Number of eigenvalues per q/k-point."""
        if "Vibrational" in self.__class__.__name__:
            return self._data["num_modes"]
        return self._data["num_bands"]

    @property
    def projector_weights(self):
        """Return the array of projector weights per eigval, with shape
        (num_projectors, num_kpoints, num_bands).

        """
        return self._data.get("projector_weights")

    @property
    def num_projectors(self):
        """Return the number of projectors."""
        if self.projectors is None:
            return 0
        return len(self.projectors)

    @property
    def kpoint_branches(self):
        """Return the k-point branches in the older format, which
        contained a list of lists of continous indices.

        """
        if self._data.get("kpoint_branches") is None:
            self._data["kpoint_branches"] = self.find_full_kpt_branch()
        return self._data["kpoint_branches"]

    @property
    def kpoint_branch_start(self):
        """Return the indices of the start of branches."""
        if not self._data.get("kpoint_branch_start"):
            self.set_branches_and_spacing()
        return self._data["kpoint_branch_start"]

    @property
    def kpoint_path_spacing(self):
        """An estimated kpoint spacing."""
        if not self._data.get("kpoint_path_spacing"):
            self.set_branches_and_spacing()
        return self._data["kpoint_path_spacing"]

    @property
    def kpoint_path(self):
        """The fractional sampling path in reciprocal space."""
        return np.asarray(self._data["kpoint_path"])

    @property
    def kpoint_weights(self):
        if "kpoint_weights" in self._data:
            return np.asarray(self._data["kpoint_weights"])
        return None

    @property
    def kpoint_path_cartesian(self):
        """The reicprocal space sampling path in Cartesian coordinates."""
        return np.asarray(frac2cart(real2recip(self.lattice_cart), self.kpoint_path))

    @property
    def num_spins(self):
        """Dummy number of spins."""
        return 1

    @property
    def spin_fermi_energy(self):
        """Dummy Fermi energy per spin channel."""
        return [0]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The config simply loads the default or user-specified config. """


__all__ = ["load_custom_settings", "set_settings", "SETTINGS"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"

from .config import load_custom_settings, SETTINGS, set_settings



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The fingerprints module provides functionality for calculating, in
parallel, structural fingerprints like PDF and PXRD.

"""


__all__ = [
    "get_uniq_cursor",
    "Fingerprint",
    "FingerprintFactory",
    "PDF",
    "PDFOverlap",
    "CombinedProjectedPDF",
    "PXRD",
]

__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .similarity import get_uniq_cursor
from .fingerprint import Fingerprint, FingerprintFactory
from .pdf import PDF, PDFOverlap, CombinedProjectedPDF
from .pxrd import PXRD



"""This submodule implements some convenience functions for working with OPTIMADE structures."""
from typing import Dict, Any, List, Union
from matador.crystal import Crystal


def optimade2dict_from_url(url: str) -> Union[Crystal, List[Crystal]]:
    """Queries the provided OPTIMADE URL and returns
    a Crystal of list of crystals for the corresponding
    structures.

    Parameters:
        url: The URL of a single or multiple OPTIMADE structure entries.

    Returns:
        The crystal or list of crystals.

    """
    import requests

    data = requests.get(url).json()["data"]
    if isinstance(data, list):
        return [optimade2dict(data) for data in data]

    return optimade2dict(data)


def optimade2dict(structure: Dict[str, Any]) -> Crystal:
    """This function takes an OPTIMADE structure and converts
    it into a matador Crystal."""
    return Crystal(
        {
            "lattice_cart": structure["attributes"]["lattice_vectors"],
            "positions_abs": structure["attributes"]["cartesian_site_positions"],
            "atom_types": structure["attributes"]["species_at_sites"],
            "source": [structure["id"]],
            "optimade_structure": structure,
        }
    )



import unittest


class ScatteringDataLoaderTests(unittest.TestCase):
    """Test whether the atomic scattering factors can be loaded."""

    def test_load_gsas(self):
        from matador.data.atomic_scattering import GSAS_ATOMIC_SCATTERING_COEFFS

        self.assertEqual(len(GSAS_ATOMIC_SCATTERING_COEFFS["Cu"]), 3)
        for key in GSAS_ATOMIC_SCATTERING_COEFFS:
            self.assertEqual(len(GSAS_ATOMIC_SCATTERING_COEFFS[key]), 3)

    def test_load_raspa(self):
        from matador.data.atomic_scattering import RASPA_ATOMIC_SCATTERING_COEFFS

        self.assertEqual(len(RASPA_ATOMIC_SCATTERING_COEFFS["Cu"]), 3)
        for key in RASPA_ATOMIC_SCATTERING_COEFFS:
            self.assertEqual(len(RASPA_ATOMIC_SCATTERING_COEFFS[key]), 3)


class MagresDataLoaderTests(unittest.TestCase):
    """Test whether magres data can be loaded correctly."""

    def test_load_efg(self):
        from matador.data.magres import ELECTRIC_QUADRUPOLE_MOMENTS

        self.assertEqual(len(ELECTRIC_QUADRUPOLE_MOMENTS), 93)


class PeriodicTableTests(unittest.TestCase):
    """Test that the periodic table is written correctly."""

    def test_numbers(self):
        from matador.data.periodic_table import PERIODIC_TABLE

        for i in range(92):
            assert list(PERIODIC_TABLE.values())[i].number == i

    def test_keys_and_symbols(self):
        from matador.data.periodic_table import PERIODIC_TABLE

        for element in PERIODIC_TABLE:
            assert PERIODIC_TABLE[element].symbol == element



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some useful functions for real/reciprocal
cell manipulation, symmetry checking and sampling (e.g. grids and paths.)

"""

from __future__ import annotations
from typing import Dict, Any, Union, Tuple, List, TYPE_CHECKING

import numpy as np

from matador.data.periodic_table import PERIODIC_TABLE

if TYPE_CHECKING:
    from matador.crystal import Crystal


EPS = 1e-12


def abc2cart(lattice_abc: List[List[float]]) -> List[List[float]]:
    """Converts lattice parameters into Cartesian lattice vectors.

    Parameters:
        lattice_abc: The lattice parameters [[a, b, c], [alpha, beta, gamma]]

    Returns:
        Cartesian lattice vectors.

    """
    assert len(lattice_abc) == 2
    assert len(lattice_abc[0]) == 3
    assert len(lattice_abc[1]) == 3
    a = lattice_abc[0][0]
    b = lattice_abc[0][1]
    c = lattice_abc[0][2]
    deg2rad = np.pi / 180
    alpha = lattice_abc[1][0] * deg2rad
    beta = lattice_abc[1][1] * deg2rad
    gamma = lattice_abc[1][2] * deg2rad
    lattice_cart = []
    lattice_cart.append([a, 0.0, 0.0])
    bx = b * np.cos(gamma)
    by = b * np.sin(gamma)
    tol = 1e-12
    if abs(bx) < tol:
        bx = 0.0
    if abs(by) < tol:
        by = 0.0
    cx = c * np.cos(beta)
    if abs(cx) < tol:
        cx = 0.0
    cy = c * (np.cos(alpha) - np.cos(beta) * np.cos(gamma)) / np.sin(gamma)
    if abs(cy) < tol:
        cy = 0.0
    cz = np.sqrt(c**2 - cx**2 - cy**2)
    lattice_cart.append([bx, by, 0.0])
    lattice_cart.append([cx, cy, cz])
    return lattice_cart


def cart2abcstar(lattice_cart: List[List[float]]) -> np.ndarray:
    """Convert Cartesian lattice vectors to the reciprocal of the lattice vectors,
    NOT the reciprocal lattice vectors (useful when writing PDB files).

    Parameters:
        lattice_cart: Cartesian lattice vectors.

    Returns:
        Reciprocal of the lattice vectors.

    """
    return np.asarray(real2recip(lattice_cart)) / (2 * np.pi)


def cart2volume(lattice_cart: List[List[float]]) -> float:
    """Convert lattice_cart to cell volume.

    Parameters:
        lattice_cart: Cartesian lattice vectors.

    Returns:
        Cell volume in the same unit as the lattice vectors (cubed).

    """
    lattice_cart = np.asarray(lattice_cart)
    vol = np.abs(np.dot(np.cross(lattice_cart[0], lattice_cart[1]), lattice_cart[2]))
    return vol


def cart2abc(lattice_cart: List[List[float]]) -> List[List[float]]:
    """Convert Cartesian lattice vectors to lattice parametres.

    Parameters:
        lattice_cart: Cartesian lattice vectors.

    Returns:
        The lattice parameters :math:`[(a, b, c), (\\alpha, \\beta, \\gamma)]`.

    """
    vecs = lattice_cart
    lattice_abc = []
    a, b, c = (np.sqrt(sum([val**2 for val in vec])) for vec in vecs)
    lattice_abc.append([a, b, c])
    # np.cos(alpha) = b.c /|b * c|
    radians2deg = 180.0 / np.pi
    cos_alpha = sum([val_b * val_c for (val_c, val_b) in zip(vecs[2], vecs[1])]) / (
        b * c
    )
    cos_beta = sum([val_c * val_a for (val_c, val_a) in zip(vecs[2], vecs[0])]) / (
        c * a
    )
    cos_gamma = sum([val_a * val_b for (val_a, val_b) in zip(vecs[0], vecs[1])]) / (
        a * b
    )
    alpha = radians2deg * np.arccos(cos_alpha)
    beta = radians2deg * np.arccos(cos_beta)
    gamma = radians2deg * np.arccos(cos_gamma)
    lattice_abc.append([alpha, beta, gamma])
    return lattice_abc


def frac2cart(
    lattice_cart: List[List[float]], positions_frac: List[List[float]]
) -> List[List[float]]:
    """Convert positions_frac block into positions_abs.

    Parameters:
        lattice_cart: Cartesian lattice vectors.
        positions_frac: List of fractional position vectors.

    Returns:
        List of absolute position vectors.

    """
    _positions_frac = np.asarray(positions_frac)
    reshaped = False
    if len(np.shape(_positions_frac)) == 1:
        reshaped = True
        _positions_frac = _positions_frac.reshape((1, 3))
    _lattice_cart = np.asarray(lattice_cart)
    positions_abs = switch_coords(_lattice_cart, _positions_frac)
    if reshaped:
        positions_abs = positions_abs.reshape(-1)
    return positions_abs.tolist()


def wrap_frac_coords(
    positions: Union[List[List[float]], List[float]], remove: bool = False
) -> Union[List[List[float]], List[float]]:
    """Wrap the given fractional coordinates back into the cell.

    Parameters:
        positions: list of fractional position vectors, or
            a single position.

    Keyword arguments:
        remove: if True, removes points exterior to the cell.

    Returns:
        List of wrapped fractional position vectors.

    """
    from copy import deepcopy

    wrapped = deepcopy(positions)
    single = False
    if len(wrapped) == 3 and isinstance(wrapped[0], float):
        wrapped = [wrapped]
        single = True
    if remove:
        to_remove = len(wrapped) * [False]
    for ind, pos in enumerate(wrapped):
        for k in range(3):
            if pos[k] >= 1 or pos[k] < 0:
                if remove:
                    to_remove[ind] = True
                    break
                else:
                    wrapped[ind][k] %= 1
    if remove:
        wrapped = [wrapped[ind] for ind, res in enumerate(to_remove) if not res]
    if single:
        return wrapped[0]

    return wrapped


def switch_coords(
    lattice: np.ndarray, pos: np.ndarray, norm: float = None
) -> np.ndarray:
    """Act on coordinates with the relevant lattice
    vectors to switch from fractional to absolute coordinates.

    Parameters:

        lattice: either lattice_cart or reciprocal lattice_cart (3x3 array)
        pos: input positions to convert (3xN array for N atoms).

    Keyword arguments:
        norm (float): divide final coordinates by normalisation factor, e.g.
            :math:`2 \\pi` when lattice is recip and positions are cartesian.

    Returns:
        3xN array of converted positions.

    """
    new_pos = np.zeros_like(pos)
    for ni, _ in enumerate(pos):
        for j in range(3):
            new_pos[ni] += lattice[j] * pos[ni][j]
    if norm is not None:
        new_pos /= norm
    return new_pos


def cart2frac(
    lattice_cart: List[List[float]], positions_abs: List[List[float]]
) -> List[List[float]]:
    """Convert positions_abs block into positions_frac (and equivalent
    in reciprocal space).

    Parameters:
        lattice_cart: Cartesian lattice vectors.
        positions_abs: list of absolute position vectors.

    Returns:
        List of fractional position vectors with the same shape
            as the input list.

    """
    _positions_abs = np.asarray(positions_abs, dtype=np.float64)
    reshaped = False
    if len(np.shape(_positions_abs)) == 1:
        reshaped = True
        _positions_abs = _positions_abs.reshape((1, 3))
    recip_lat = np.asarray(real2recip(lattice_cart))
    recip_lat = recip_lat.T
    positions_frac = switch_coords(recip_lat, _positions_abs, norm=2 * np.pi)
    if reshaped:
        positions_frac = positions_frac.reshape(-1)
    return positions_frac.tolist()


def real2recip(real_lat: List[List[float]]) -> List[List[float]]:
    """Convert the real lattice in Cartesian basis to
    the reciprocal space lattice.

    Parameters:
        real_lat: Cartesian lattice vectors.

    Returns:
        Cartesian lattice vectors of reciprocal lattice.

    """
    real_lat = np.asarray(real_lat)
    recip_lat = np.zeros((3, 3))
    volume = np.dot(real_lat[0], np.cross(real_lat[1], real_lat[2]))
    recip_lat[0] = (2 * np.pi) * np.cross(real_lat[1], real_lat[2])
    recip_lat[1] = (2 * np.pi) * np.cross(real_lat[2], real_lat[0])
    recip_lat[2] = (2 * np.pi) * np.cross(real_lat[0], real_lat[1])
    recip_lat /= volume
    return recip_lat.tolist()


def calc_mp_grid(
    lattice_cart: List[List[float]], spacing: float
) -> Tuple[int, int, int]:
    """Return the Monkhorst-Pack grid based on lattice
    vectors and desired spacing.

    Parameters:
        lattice_cart: Cartesian lattice vectors.
        spacing: desired maximum grid spacing.

    Returns:
        List of 3 integers defining the MP grid.

    """
    recip_lat = real2recip(lattice_cart)
    recip_len = np.zeros((3))
    recip_len = np.sqrt(np.sum(np.power(recip_lat, 2), axis=1))
    mp_grid = recip_len / (2 * np.pi * spacing)
    return [int(np.ceil(elem)) for elem in mp_grid]


def shift_to_include_gamma(mp_grid: Tuple[int, int, int]) -> Tuple[float, float, float]:
    """Calculate the shift required to include the :math:`\\Gamma`-point
    in the Monkhorst-Pack grid.

    Parameters:
        mp_grid: number of grid points
            in each reciprocal space direction.

    Returns:
        The shift required to include the :math:`\\Gamma`.

    """
    shift = [0, 0, 0]
    for ind, val in enumerate(mp_grid):
        if val % 2 == 0:
            shift[ind] = 1.0 / (val * 2)
    return shift


def shift_to_exclude_gamma(mp_grid: Tuple[int, int, int]) -> Tuple[float, float, float]:
    """Calculate the shift required to exclude the :math:`\\Gamma`-point
    in the Monkhorst-Pack grid. Returns the "minimal shift", i.e. only
    one direction will be shifted.

    Parameters:
        mp_grid: number of grid points
            in each reciprocal space direction.

    Returns:
        The shift required to exclude :math:`\\Gamma`.

    """
    shift = [0, 0, 0]
    if all([val % 2 == 1 for val in mp_grid]):
        for ind, val in enumerate(mp_grid):
            if val % 2 == 1:
                shift[ind] = 1.0 / (val * 2)
                break

    return shift


def get_best_mp_offset_for_cell(doc: Union[Dict[str, Any], Crystal]) -> List[float]:
    """Calculates the "best" kpoint_mp_offset to use for the passed
    cell. If the crystal has a hexagonal space group, then the offset
    returned will shift the grid to include :math:`\\Gamma`-point, and vice
    versa for non-hexagonal cells.

    Parameters:
        doc: matador document/Crystal to consider, containing structural
            information and a "kpoints_mp_spacing" key.

    Returns:
        The desired `kpoint_mp_offset`.

    """

    gamma = False
    if "6" in get_spacegroup_spg(doc, symprec=1e-5):
        gamma = True

    if "lattice_cart" not in doc or "kpoints_mp_spacing" not in doc:
        raise RuntimeError("Unable to calculate offset without lattice or spacing")

    mp_grid = calc_mp_grid(doc["lattice_cart"], doc["kpoints_mp_spacing"])
    if gamma:
        return shift_to_include_gamma(mp_grid)

    return shift_to_exclude_gamma(mp_grid)


def calc_mp_spacing(
    real_lat: List[List[float]], mp_grid: Tuple[int, int, int], prec: int = 3
) -> float:
    """Convert real lattice in Cartesian basis and the CASTEP
    `kpoint_mp_grid` into a CASTEP grid spacing.

    Parameters:
        real_lat: Cartesian lattice vectors.
        mp_grid: 3 integers defining the MP grid.

    Keyword arguments:
        prec: desired decimal precision of output.

    Returns:
        `kpoint_mp_spacing` rounded to `prec` decimal places.

    """
    recip_lat = real2recip(real_lat)
    recip_len = np.zeros((3))
    recip_len = np.sqrt(np.sum(np.power(recip_lat, 2), axis=1))
    spacing = recip_len / (2 * np.pi * np.asarray(mp_grid))
    max_spacing = np.max(spacing)
    exponent = round(np.log10(max_spacing) - prec)
    return round(max_spacing + 0.5 * 10**exponent, prec)


def get_seekpath_kpoint_path(
    doc: Union[Dict[str, Any], Tuple],
    standardize: bool = True,
    explicit: bool = True,
    spacing: float = 0.01,
    threshold: float = 1e-7,
    debug: bool = False,
    symmetry_tol: float = None,
) -> Tuple[Dict[str, Any], List[List[float]], Dict[str, Any]]:
    """Return the conventional kpoint path of the relevant crystal system
    according to the definitions by "HKPOT" in
    Comp. Mat. Sci. 128, 2017:

    http://dx.doi.org/10.1016/j.commatsci.2016.10.015

    Parameters:
        doc: matador doc or spglib tuple to find kpoint path for.

    Keyword arguments:
        spacing: desired kpoint spacing
        threshold: internal seekpath threshold
        symmetry_tol: spglib symmetry tolerance

    Returns:
        dict: standardized version of input doc
        list: list of kpoint positions
        dict: full dictionary of all seekpath results

    """
    try:
        from seekpath import get_explicit_k_path, get_path
    except ImportError:
        raise ImportError(
            "SeeK-Path dependency missing, please install it with `pip install seekpath`."
        )

    if symmetry_tol is None:
        symmetry_tol = 1e-5

    if isinstance(doc, tuple):
        spg_structure = doc
    else:
        if standardize:
            spg_structure = doc2spg(standardize_doc_cell(doc, symprec=symmetry_tol))
        else:
            spg_structure = doc2spg(doc)

    if explicit:
        seekpath_results = get_explicit_k_path(
            spg_structure,
            reference_distance=spacing,
            with_time_reversal=True,
            symprec=symmetry_tol,
            threshold=threshold,
        )

        kpt_path = seekpath_results["explicit_kpoints_rel"]
    else:
        seekpath_results = get_path(spg_structure)
        kpt_path = []

    primitive_doc = dict()
    primitive_doc["lattice_cart"] = seekpath_results["primitive_lattice"]
    primitive_doc["positions_frac"] = seekpath_results["primitive_positions"]
    primitive_doc["atom_types"] = [
        list(PERIODIC_TABLE)[i] for i in seekpath_results["primitive_types"]
    ]
    primitive_doc["num_atoms"] = len(primitive_doc["atom_types"])
    primitive_doc["lattice_abc"] = cart2abc(primitive_doc["lattice_cart"])
    primitive_doc["cell_volume"] = cart2volume(primitive_doc["lattice_cart"])
    if debug:
        print(
            "Found lattice type {}".format(seekpath_results["bravais_lattice_extended"])
        )
        print("Old lattice:\n", np.asarray(doc["lattice_cart"]))
        print("Contained {} atoms".format(doc["num_atoms"]))
        print("New lattice:\n", np.asarray(primitive_doc["lattice_cart"]))
        print("Contains {} atoms".format(primitive_doc["num_atoms"]))
        print("k-point path contains {} points.".format(len(kpt_path)))

    if "site_occupancy" in doc:
        if min(doc["site_occupancy"]) < 1 - EPS:
            print("Ignoring any site occupancy found in this cell.")
        primitive_doc["site_occupancy"] = [1 for atom in primitive_doc["atom_types"]]

    return primitive_doc, kpt_path, seekpath_results


def doc2spg(
    doc: Union[Dict[str, Any], "Crystal"], check_occ: bool = True
) -> Tuple[List[List[float]], List[List[float]], List[int]]:
    """Return an spglib input tuple from a matador doc.

    Parameters:
        doc: matador document or Crystal object.

    Keyword arguments:
        check_occ: check for partial occupancy and raise an error if present.

    Returns:
        tuple: spglib-style tuple of lattice, positions and types.

    """
    from matador.utils.chem_utils import get_atomic_number

    try:
        if "lattice_cart" not in doc:
            doc["lattice_cart"] = abc2cart(doc["lattice_abc"])
    except KeyError:
        raise RuntimeError("doc2spg failed: unable to find lattice in document.")

    required_keys = ("lattice_cart", "positions_frac", "atom_types")
    if all(key in doc for key in required_keys):
        if "lattice_cart" not in doc:
            doc["lattice_cart"] = abc2cart(doc["lattice_abc"])
        cell = (
            doc["lattice_cart"],
            doc["positions_frac"],
            [get_atomic_number(elem) for elem in doc["atom_types"]],
        )
        if check_occ and np.min(doc.get("site_occupancy", [1.0])) < 1.0:
            raise RuntimeError("spglib does not support partial occupancy.")

        return cell

    else:
        raise RuntimeError(
            f"Unable to use doc2spg, one of {required_keys} was missing."
        )


def get_space_group_label_latex(label: str) -> str:
    """Return the LaTeX format of the passed space group label. Takes
    any string, leaves the first character upright, italicses the rest,
    handles subscripts and bars over numbers.

    Parameters:
        label: a given space group in "standard" plain text format,
        e.g. P-63m to convert to '$P\\bar{6}3m$'.

    Returns:
        The best attempt to convert the label to LaTeX format.

    """
    import re

    return "${}$".format(re.sub("-(?P<number>[0-9])", "\\\\bar{\\g<number>}", label))


def standardize_doc_cell(
    doc: Union["Crystal", Dict[str, Any]], primitive: bool = True, symprec: float = 1e-2
) -> Union["Crystal", Dict[str, Any]]:
    """Return standardized cell data from matador doc.

    Parameters:
        doc: matador document or Crystal object to standardize.

    Keyword arguments:
        primitive: whether to reduce cell to primitive.
        symprec: spglib symmetry tolerance.

    Returns:
        A matador document/`Crystal` containing standardized cell.

    """
    import spglib as spg
    from matador.utils.chem_utils import get_atomic_symbol
    from copy import deepcopy
    from matador.crystal import Crystal

    spg_cell = doc2spg(doc)
    spg_standardized = spg.standardize_cell(
        spg_cell, to_primitive=primitive, symprec=symprec
    )
    if not isinstance(doc, Crystal):
        std_doc = deepcopy(doc)
    else:
        std_doc = deepcopy(doc._data)
    std_doc["lattice_cart"] = [list(vec) for vec in spg_standardized[0]]
    std_doc["lattice_abc"] = cart2abc(std_doc["lattice_cart"])
    std_doc["positions_frac"] = [list(atom) for atom in spg_standardized[1]]
    std_doc["atom_types"] = [get_atomic_symbol(atom) for atom in spg_standardized[2]]
    std_doc["site_occupancy"] = len(std_doc["positions_frac"]) * [1]
    std_doc["cell_volume"] = cart2volume(std_doc["lattice_cart"])
    std_doc["space_group"] = get_spacegroup_spg(std_doc, symprec=symprec)
    # if the original document was a crystal, return a new one
    if isinstance(doc, Crystal):
        std_doc = Crystal(std_doc)

    return std_doc


def get_spacegroup_spg(
    doc: Union[Dict[str, Any], Crystal], symprec: float = 0.01, check_occ: bool = True
):
    """Return spglib spacegroup for a cell.

    Parameters:
        doc: matador document or Crystal object.

    Keyword arguments:
        symprec: spglib symmetry tolerance.

    Returns:
        The H-M space group symbol of structure.

    """
    import spglib as spg

    spg_cell = doc2spg(doc, check_occ=check_occ)
    space_group = spg.get_spacegroup(spg_cell, symprec=symprec)
    if space_group is None:
        raise RuntimeError("Spglib was unable to calculate space group.")

    return space_group.split(" ")[0]


def get_compatible_spacegroups(
    doc: Union[Dict[str, Any], Crystal], symprec_range=(-5, 0)
) -> Dict[float, str]:
    """Return the space group of a given crystal for the range of symprecs.

    Parameters:
        doc: The crystal to analyse.
        symprec_range: The range of symprecs to test (log space).

    Returns:
        A mapping from symprec to space group symbol.

    """
    spgs = {}
    for symprec in np.logspace(*symprec_range, num=10):
        spgs[symprec] = get_spacegroup_spg(doc, symprec=symprec)

    return spgs


def add_noise(doc: Dict[str, Any], amplitude: float = 0.1) -> Dict[str, Any]:
    """Add random noise to the positions of structure contained in doc.
    Useful for force convergence tests.

    Parameters:
        doc: dictionary containing matador structure.

    Keyword arguments:
        amplitude: maximum amplitude of noise vector.

    Raises:
        KeyError if (`lattice_cart` and `positions_frac`) or `positions_abs`
            are missing.

    Returns:
        The randomised structure.
    """
    poscart = np.asarray(
        doc.get("positions_abs")
        or frac2cart(doc["lattice_cart"], doc["positions_frac"])
    )
    for atom in poscart:
        noise_vector = np.random.rand(3)
        noise_vector /= np.sqrt(np.sum(noise_vector**2))
        noise_vector *= amplitude * (np.random.rand())
        atom += noise_vector
    doc["positions_abs"] = poscart.tolist()
    doc["positions_frac"] = cart2frac(doc["lattice_cart"], doc["positions_abs"])

    return doc


def calc_pairwise_distances_pbc(
    poscart,
    images,
    lattice,
    rmax,
    poscart_b=None,
    compress=False,
    debug=False,
    filter_zero=False,
    per_image=False,
):
    """Calculate PBC distances with SciPy's cdist, given the
    image cell vectors.

    Parameters:
        poscart (numpy.ndarray): list or array of absolute atomic coordinates.
        images: iterable of lattice vector multiples (e.g. [2, -1, 3])
            required to obtain the translation to desired image cells.
        lattice (:obj:`list` if :obj:`list`): list of lattice vectors of
            the real cell.
        rmax (float): maximum value after which to mask the array.

    Keyword arguments:
        poscart_b (numpy.ndarray): absolute positions of another type of
            atom, where only A-B distances will be calculated.
        debug (bool): print timing data and how many distances were masked.
        compress (bool): whether or not to compressed the output array,
            useful when e.g. creating PDFs but not when atom ID is important.
        filter_zero (bool): whether or not to filter out the "self-interaction"
            zero distances.
        per_image (bool): return a list of distances per image, as opposed to
            one large flat. This preserves atom IDs for use elsewhere.

    Returns:
        distances (numpy.ndarray): pairwise 2-D d_ij masked array with values
            or stripped 1-D array containing just the distances, or a list of
            numpy arrays if per_image is True.

    """
    from scipy.spatial.distance import cdist
    import time

    if debug:
        start = time.time()
    _lattice = np.asarray(lattice)
    _poscart = np.asarray(poscart)
    image_distances = []
    if poscart_b is None:
        _poscart_b = _poscart
    else:
        _poscart_b = np.asarray(poscart_b)

    num_pairs = len(_poscart) * len(_poscart_b)

    if per_image:
        for image_ind, prod in enumerate(images):
            distances = cdist(_poscart, _poscart_b + prod @ _lattice)
            distances = np.ma.masked_where(distances > rmax, distances, copy=False)
            image_distances.append(distances)

        distances = image_distances

    else:
        array_size = len(_poscart) * len(images) * len(_poscart_b)
        distances = np.empty(array_size)
        for image_ind, prod in enumerate(images):
            distances[image_ind * num_pairs : (image_ind + 1) * num_pairs] = cdist(
                _poscart, _poscart_b + prod @ _lattice
            ).flatten()

        distances = np.ma.masked_where(distances > rmax, distances, copy=False)
        if filter_zero:
            distances = np.ma.masked_where(distances < EPS, distances, copy=False)

        if debug:
            print(
                "Calculated: {}, Used: {}, Ignored: {}".format(
                    len(distances),
                    np.ma.count(distances),
                    np.ma.count_masked(distances),
                )
            )
        if compress:
            distances = distances.compressed()

    if debug:
        end = time.time()
        print("Calculated distances in {} s".format(end - start))

    return distances


def create_simple_supercell(
    doc: Union[Dict[str, Any], "Crystal"],
    extension: Tuple[int, int, int],
    standardize: bool = False,
    symmetric: bool = False,
) -> Union[Dict[str, Any], "Crystal"]:
    """Return a document with new supercell, given extension vector.

    Parameters:
        doc: matador doc to construct cell from.
        extension: multiplicity of each lattice vector, e.g. (2,2,1).

    Keyword arguments:
        standardize: whether or not to use spglib to standardize the cell first.
        symmetric: whether or not centre the new cell on the origin.

    Returns:
        The supercell, either as a `Crystal` or as a dictionary,
        depending on the input type.

    """
    from itertools import product
    from copy import deepcopy
    from matador.crystal import Crystal

    if not all([elem >= 1 for elem in extension]) or extension == (1, 1, 1):
        raise RuntimeError("Bad/null supercell {} requested...".format(extension))

    if isinstance(doc, Crystal):
        supercell_doc = deepcopy(doc._data)
    else:
        supercell_doc = deepcopy(doc)

    # copy new doc and delete data that will not be corrected
    keys_to_copy = set(
        ["positions_frac", "lattice_cart", "atom_types", "source", "stoichiometry"]
    )
    supercell_doc = {key: deepcopy(doc[key]) for key in keys_to_copy}

    # standardize cell with spglib
    if standardize:
        supercell_doc = standardize_doc_cell(supercell_doc)

    for i, elem in enumerate(extension):
        supercell_doc["lattice_cart"] = np.asarray(supercell_doc["lattice_cart"])
        for k in range(3):
            supercell_doc["lattice_cart"][i][k] = (
                elem * supercell_doc["lattice_cart"][i][k]
            )
        for ind, atom in enumerate(supercell_doc["positions_frac"]):
            supercell_doc["positions_frac"][ind][i] /= elem

    images = product(*[list(range(elem)) for elem in extension])
    new_positions = []
    new_atoms = []
    for image in images:
        if image != (0, 0, 0):
            for ind, atom in enumerate(supercell_doc["atom_types"]):
                new_pos = deepcopy(supercell_doc["positions_frac"][ind])
                for i, elem in enumerate(image):
                    new_pos[i] += image[i] / extension[i]
                new_positions.append(new_pos)
                new_atoms.append(atom)
    supercell_doc["atom_types"].extend(new_atoms)
    supercell_doc["positions_frac"].extend(new_positions)

    if symmetric:
        supercell_doc["positions_frac"] = np.asarray(supercell_doc["positions_frac"])
        supercell_doc["positions_frac"] -= np.mean(
            supercell_doc["positions_frac"], axis=0
        )
        supercell_doc["positions_frac"] = supercell_doc["positions_frac"].tolist()

    supercell_doc["num_atoms"] = len(supercell_doc["atom_types"])
    supercell_doc["cell_volume"] = cart2volume(supercell_doc["lattice_cart"])
    supercell_doc["lattice_abc"] = cart2abc(supercell_doc["lattice_cart"])
    supercell_doc["source"] = doc["source"]
    supercell_doc["stoichiometry"] = doc["stoichiometry"]

    assert np.isclose(
        supercell_doc["cell_volume"],
        np.prod(extension) * cart2volume(doc["lattice_cart"]),
    )

    if isinstance(doc, Crystal):
        return Crystal(supercell_doc)

    return supercell_doc


def create_supercell_with_minimum_side_length(
    doc: Union[Dict[str, Any], "Crystal"], target: float
):
    """Pad the cell such that the minimum side length is greater than the target length.

    Parameters:
        doc: The crystal structure to pad.
        target: The target minimum side length.

    Returns:
        The supercell.

    """
    extension = [1, 1, 1]
    for ind, lat in enumerate(doc.cell.lengths):
        if lat < target:
            extension[ind] = int(-(target // -lat))

    return create_simple_supercell(doc, extension, standardize=False)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements some useful classes for manipulating
DOS and dispersion data.

"""

from .dos import VibrationalDOS, ElectronicDOS, DensityOfStates
from .dispersion import VibrationalDispersion, ElectronicDispersion, Dispersion
from .spectral import Spectral


__all__ = [
    "VibrationalDOS",
    "ElectronicDOS",
    "DensityOfStates",
    "VibrationalDispersion",
    "ElectronicDispersion",
    "Dispersion",
    "Spectral",
]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements an interface for querying the
__changelog_<collection> collections to allow for display and reversion
of particular database changes.

"""


from matador.db import make_connection_to_collection
from matador.utils.print_utils import print_warning, print_notify


class DatabaseChanges:
    """Class to view and undo particular
    database changesets.

    """

    def __init__(
        self,
        collection_name: str,
        changeset_ind=0,
        action="view",
        override=False,
        mongo_settings=None,
    ):
        """Parse arguments and run changes interface.

        Parameters:
            collection_name (str): the base collection name to act upon

        Keyword arguments:
            changset_ind (int): the number of the changset to act upon (1 is oldest)
            action (str): either 'view' or 'undo'
            override (bool): override all options to positive answers for testing
            mongo_settings (dict): dictionary of already-sources mongo settings

        """
        self.changelog_name = "__changelog_{}".format(collection_name)
        _, _, self.collections = make_connection_to_collection(
            self.changelog_name,
            allow_changelog=True,
            override=override,
            mongo_settings=mongo_settings,
        )
        self.repo = [self.collections[key] for key in self.collections][0]
        curs = list(self.repo.find())

        if not curs:
            exit("No changesets found for {}".format(collection_name))

        # if no changeset specified, print summary
        if changeset_ind == 0:
            self.print_change_summary(curs)

        elif changeset_ind > len(curs):
            exit(
                'No changeset {} found for collection called "{}".'.format(
                    changeset_ind, collection_name
                )
            )

        # otherwise, try to act on particular changeset
        elif changeset_ind <= len(curs):
            self.change = curs[changeset_ind - 1]
            self.view_changeset(self.change, changeset_ind - 1)
            if action == "undo":
                count = curs[changeset_ind - 1]["count"]
                print_warning(
                    "An attempt will now be made to remove {} structures from {}.".format(
                        count, collection_name
                    )
                )
                print_notify("Are you sure you want to do that? (y/n)")
                if override:
                    response = "y"
                else:
                    response = input()
                if response.lower() == "y":
                    print_notify("You don't have any doubts at all? (y/n)")
                    if override:
                        next_response = "n"
                    else:
                        next_response = input()
                    if next_response.lower() == "n":
                        print("You're the boss, deleting structures now...")
                    else:
                        exit("As I thought...")
                else:
                    exit()

                # proceed with deletion
                _, _, collections = make_connection_to_collection(
                    collection_name, allow_changelog=False, override=override
                )
                collection_to_delete_from = [collections[key] for key in collections][0]
                result = collection_to_delete_from.delete_many(
                    {"_id": {"$in": self.change["id_list"]}}
                )
                print(
                    "Deleted {}/{} successfully.".format(
                        result.deleted_count, self.change["count"]
                    )
                )
                print("Tidying up changelog database...")
                self.repo.delete_one({"_id": self.change["_id"]})
                if not self.repo.find_one():
                    print("No structures left remaining, deleting database...")
                    collection_to_delete_from.drop()
                    self.repo.drop()
                print("Success!")

    @staticmethod
    def view_changeset(changeset, index):
        """Prints all details about a particular changeset.

        Parameters:
            changeset (dict): changeset stored in changelog database
            index (int): changeset index

        """
        print("Files added by changeset:")
        for src in changeset["src_list"]:
            print("(+) {}".format(src))
        print(
            "({}) {} {:>7d} additions".format(
                index + 1, changeset["date"], changeset["count"]
            )
        )

    @staticmethod
    def print_change_summary(curs):
        """Prints a summary of changes.

        Parameters:
            curs (list): cursor from changelog database

        """
        for index, change in enumerate(curs):
            print(
                "({}) {} {:>7d} additions".format(
                    index + 1, change["date"], change["count"]
                )
            )



# coding: utf-8
# Distributed under the terms of the MIT license.

from typing import Dict, List, Optional, Tuple
import warnings

import numpy as np
import statsmodels.api as sm

from matador.plotting.plotting import plotting_function
from matador.crystal import Crystal

__all__ = ("MagresReferencer",)


class MagresReferencer:
    """Class for referencing computed NMR chemical shielding tensors
    with experimental data on chemical shifts, and related plotting.

    Attributes:
        fit_model: The underlying statsmodel model.
        fit_results: The results of the statsmodel fit.
        fit_intercept: The intercept of the fit.
        fit_gradient: The gradient of the fit.
        fit_rsquared: The R^2 value of the fit.
        structures_exp: A dictionary of experimental structures, keyed by label
        shifts_exp: A dictionary of measured shifts with the same keys as `structures_exp`.
        species: The species of interest.
        structures: An optional list of theoretical structures with computed shieldings
            to be referenced.
        warn_tolerance: The maximum deviation from the ideal "-1" gradient above which to warn
            the user during the fit.

    """

    fit_model: sm.regression.linear_model.WLS = None
    fit_results: sm.regression.linear_model.RegressionResults = None
    fit_intercept: float
    fit_gradient: float
    fit_rsquared: float
    structures_exp: Dict[str, Crystal]
    shifts_exp: List[Dict[str, List[float]]]
    species: str
    structures: Optional[List[Crystal]]
    warn_tolerance: float = 0.1

    def __init__(
        self,
        structures_exp: Dict[str, Crystal],
        shifts_exp: List[Dict[str, List[float]]],
        species: str,
        structures: Optional[List[Crystal]] = None,
        warn_tolerance: float = 0.1,
    ):
        self.structures_exp = structures_exp
        self.shifts_exp = shifts_exp
        self.species = species
        self.structures = structures
        self.warn_tolerance = warn_tolerance

        self._calc_shifts = []
        self._expt_shifts = []
        self._fit_weights = []
        self._fit_structures = []
        self._fitted = False

        for formula in self.shifts_exp:
            if self.species not in self.shifts_exp[formula]:
                continue

            if formula not in self.structures_exp:
                warnings.warn(f"Missing {formula} in reference structures.")

            self.match_exp_structure_shifts(
                self.structures_exp[formula], self.shifts_exp[formula][self.species]
            )

        self.fit()
        self.print_fit_summary()

        if self.structures is not None:
            self.set_shifts_from_fit(self.structures)

    def match_exp_structure_shifts(self, structure, shifts):
        """For a model structure and a set of experimental shifts, match each site in the
        structure to the closest shift value, allowing for multiplicity.

        Sets the weights to be used for each site based on the number of unique shifts/local
        environments in the structure.

        Parameters:
            structure: The model structure for each exp. shift.
            shifts: An array of measured chemical shifts.

        """
        relevant_sites = [site for site in structure if site.species == self.species]
        calc_shifts = sorted(
            [site["chemical_shielding_iso"] for site in relevant_sites]
        )
        _shifts = shifts
        if (
            len(_shifts) <= len(relevant_sites)
            and len(relevant_sites) % len(_shifts) == 0
        ):
            multiplier = len(relevant_sites) // len(_shifts)
            _shifts = [shift for cell in [_shifts] * multiplier for shift in cell]
            _weights = [1 / multiplier for shift in _shifts]
        else:
            raise RuntimeError(
                f"Incompatible shift sizes: {len(relevant_sites)} (theor.) vs {len(_shifts)} (expt.), "
                "please provide commensurate cells."
            )

        _shifts = sorted(_shifts, reverse=True)

        self._calc_shifts.extend(calc_shifts)
        self._expt_shifts.extend(_shifts)
        self._fit_weights.extend(_weights)
        self._fit_structures.extend(len(_shifts) * [structure.formula_tex])

        return _shifts, _weights, calc_shifts

    def set_shifts_from_fit(self, structures):
        """Set the chemical shifts of the given structures to the predicted values.

        Parameters:
            structures: A list of structures with computed shieldings.

        """
        for _, struc in enumerate(structures):
            for _, site in enumerate(struc):
                if site.species == self.species:
                    site["chemical_shift_iso"] = self.predict(
                        site["chemical_shielding_iso"]
                    )[0]

    def fit(self):
        """Construct a statsmodels weighted least squares model between experimental
        shifts and computed shieldings.

        Sets the following attributes:
            fit_model: The underlying statsmodel model.
            fit_results: The output of the statsmodel fit.
            fit_intercept: The intercept of the fit.
            fit_gradient: The gradient of the fit.
            fit_rsquared: The R^2 value of the fit.

        """
        _fit_shifts = sm.add_constant(self._calc_shifts)
        self.fit_model = sm.regression.linear_model.WLS(
            self._expt_shifts, _fit_shifts, weights=self._fit_weights
        )
        self.fit_results = self.fit_model.fit()
        self.fit_intercept = self.fit_results.params[0]
        self.fit_gradient = self.fit_results.params[1]
        self.fit_rsquared = self.fit_results.rsquared
        self._fitted = True

        if abs(self.fit_gradient + 1.0) > self.warn_tolerance:
            warnings.warn(
                f"{self.__class__.__name__} fit gradient was {self.fit_gradient:.2f}, "
                f"outside of tolerated range -1.0 ± {self.warn_tolerance:.2f}"
            )

    def predict(self, shieldings) -> Tuple[np.array, np.array]:
        """Compute the predicted chemical shifts (and errors) for given chemical shieldings, based
        on the linear fit.

        Returns:
            A tuple of shift values and associated errors.

        """
        _shieldings = np.asarray(shieldings)
        return (
            self.fit_gradient * _shieldings + self.fit_intercept,
            self.fit_results.bse[1] * _shieldings + self.fit_results.bse[0],
        )

    def print_fit_summary(self):
        """Print the fitted parameters and their errors."""
        if self._fitted:
            print("Performed WLS fit for: δ_expt = m * σ_calc + c")
            print(f"m = {self.fit_gradient:3.3f} ± {self.fit_results.bse[1]:3.3f}")
            print(f"c = {self.fit_intercept:3.3f} ± {self.fit_results.bse[0]:3.3f} ppm")
            print(f"R² = {self.fit_rsquared:3.3f}.")
        else:
            raise RuntimeError("Fit has not yet been performed.")

    @plotting_function
    def plot_fit(self, ax=None, padding=100, figsize=None):
        """Plot the fit of shifts vs shieldings to experimental data.

        Parameters:
            ax (matplotlib.axes.Axes): Axes to plot on.
            padding (float): Padding to add to each axis limit.
            figsize (tuple): Figure size.

        Returns:
            The axis object used for plotting.

        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        ax.grid(False)
        ax.set_xlim(
            np.min(self._calc_shifts) - padding, np.max(self._calc_shifts) + padding
        )
        ax.set_ylim(
            np.min(self._expt_shifts) - padding, np.max(self._expt_shifts) + padding
        )
        ax = sns.regplot(
            y=self._expt_shifts,
            x=self._calc_shifts,
            scatter=False,
            ax=ax,
            line_kws={"linewidth": 0},
            color="grey",
            truncate=False,
        )

        sns.scatterplot(
            y=self._expt_shifts,
            x=self._calc_shifts,
            hue=self._fit_structures,
            palette="Dark2",
            ax=ax,
            zorder=100,
        )

        ax.plot(
            np.asarray(ax.get_xlim()),
            self.fit_gradient * np.asarray(ax.get_xlim()) + self.fit_intercept,
            zorder=10,
            lw=1.5,
            c="grey",
        )
        ax.legend(ncol=len(set(self._fit_structures)) // 5)
        ax.set_ylabel("$\\delta_\\mathrm{expt}$ (ppm)")
        ax.set_xlabel("$\\sigma_\\mathrm{calc}$ (ppm)")
        ax.text(
            0.1,
            0.1,
            f"$m = {self.fit_gradient:3.3f}; c = {self.fit_intercept:3.0f}\\,ppm; R^2 = {self.fit_rsquared:3.3f}$",
            bbox=dict(alpha=0.8, facecolor="w", boxstyle="Round"),
            transform=ax.transAxes,
        )

        return ax

    @plotting_function
    def plot_fit_and_predictions(self, ax=None, padding=100):
        """Make a plot of the fit and predictions for the experimental chemical shifts.

        Parameters:
            ax (matplotlib.axes.Axes): Optional axis object to plot on.
            padding (float): Padding to add to each axis limit.

        """

        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots()
        self.plot_fit(ax=ax, padding=padding)

        for doc in self.structures:
            _calc_shifts = [
                site["chemical_shielding_iso"]
                for site in doc
                if site.species == self.species
            ]
            _predicted_shifts, _predicted_errs = self.predict(_calc_shifts)
            ax.scatter(_predicted_shifts, _calc_shifts, s=5, c="k")
            ax.errorbar(
                _predicted_shifts,
                _calc_shifts,
                fmt="None",
                xerr=_predicted_errs,
                lw=0.5,
                c="k",
            )

        return ax



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some light wrappers to
the Atomic Simulation Environment (ASE).

"""

import copy
from typing import Union
from matador.utils.chem_utils import get_stoich
from matador.utils.cell_utils import get_spacegroup_spg
from matador.crystal import Crystal

__all__ = ["ase2dict", "doc2ase"]


def ase2dict(atoms, as_model=False) -> Union[dict, Crystal]:
    """Return a matador document (dictionary or :obj:`Crystal`)
    from an `ase.Atoms` object.

    Parameters:
        atoms (ase.Atoms): input structure.

    Keyword arguments:
        as_model (bool): if `True`, return a
            Crystal instead of a dictionary.

    Returns:
        Union[dict, Crystal]: matador output.

    """
    from matador.utils.cell_utils import cart2abc

    doc = {}

    # sort atoms, then their positions
    doc["atom_types"] = atoms.get_chemical_symbols()
    inds = [i[0] for i in sorted(enumerate(doc["atom_types"]), key=lambda x: x[1])]
    doc["positions_frac"] = atoms.get_scaled_positions().tolist()
    doc["positions_frac"] = [doc["positions_frac"][ind] for ind in inds]
    doc["atom_types"] = [doc["atom_types"][ind] for ind in inds]
    try:
        doc["lattice_cart"] = atoms.get_cell().tolist()
    except AttributeError:
        doc["lattice_cart"] = atoms.get_cell().array.tolist()
    doc["lattice_abc"] = cart2abc(doc["lattice_cart"])
    doc["num_atoms"] = len(doc["atom_types"])
    doc["stoichiometry"] = get_stoich(doc["atom_types"])
    doc["cell_volume"] = atoms.get_volume()
    doc["elems"] = {atom for atom in doc["atom_types"]}
    doc["num_fu"] = doc["num_atoms"] / int(
        sum(doc["stoichiometry"][i][1] for i in range(len(doc["stoichiometry"])))
    )
    doc["space_group"] = get_spacegroup_spg(doc, symprec=0.001)

    if hasattr(atoms, "info"):
        doc["ase_info"] = copy.deepcopy(atoms.info)

    if as_model:
        doc = Crystal(doc)

    return doc


def doc2ase(doc: Union[dict, Crystal], add_keys_to_info=True):
    """Convert matador document to simple ASE object.

    Parameters:
        doc (dict/:obj:`Crystal`): matador document or
            `Crystal` containing the structure.

    Keyword arguments:
        add_keys_to_info (bool): whether or not to add the keys from the
            matador document to the info section of the Atoms object.

    """
    from ase import Atoms

    atoms = Atoms(
        symbols=doc["atom_types"],
        scaled_positions=doc["positions_frac"],
        cell=doc["lattice_cart"],
        pbc=True,
    )

    if add_keys_to_info:
        if isinstance(doc, Crystal):
            atoms.info["matador"] = doc._data
        else:
            atoms.info["matador"] = copy.deepcopy(doc)
        if "_id" in atoms.info["matador"]:
            atoms.info["matador"]["_id"] = str(atoms.info["matador"]["_id"])

    return atoms



""" This file implements turning matador Crystal objects
into CrystalGraph objects.
"""
import networkx as nx
import numpy as np
import itertools

EPS = 1e-12


class CrystalGraph(nx.MultiDiGraph):
    def __init__(
        self,
        structure=None,
        graph=None,
        coordination_cutoff=1.1,
        bond_tolerance=1e20,
        num_images=1,
        debug=False,
        separate_images=False,
        delete_one_way_bonds=False,
        max_bond_length=5,
    ):
        """Create networkx.MultiDiGraph object with extra functionality for atomic networks.

        Keyword Arguments:
            structure (matador.Crystal):  crystal structure to network-ify
            graph (nx.MultiDiGraph): initialise from graph
            coordination_cutoff (float) : max multiplier of first
                coordination sphere for edge drawing
            num_images (int): number of periodic images to include in
                each direction
            separate_images (bool): whether or not to include image
                atoms as new nodes

        """

        super().__init__()

        if graph is None and structure is None:
            raise RuntimeError("No structure or graph to initialise network from.")

        if structure is not None:
            atoms = structure.sites
            num_atoms = len(atoms)
            element_bonds = {}
            images = list(
                itertools.product(range(-num_images, num_images + 1), repeat=3)
            )
            image_number = 0

            # now loop over pairs of atoms and decide whether to draw an edge
            from matador.utils.cell_utils import calc_pairwise_distances_pbc

            distances = calc_pairwise_distances_pbc(
                structure.positions_abs,
                images,
                structure.lattice_cart,
                max_bond_length,
                compress=False,
                debug=True,
            )

            # first over loop all pairs to find the minimum distance between all species pairs
            # and the minimum distance for each atom
            for i, atom in enumerate(atoms):
                self.add_node(i, species=atom.species)

            min_dists = [1e20 for atom in atoms]
            for index in np.where(~distances.mask)[0]:
                image_index = int(index / num_atoms**2)
                i = int((index - image_index * num_atoms**2) / num_atoms)
                j = int((index - image_index * num_atoms**2) % num_atoms)
                atom = atoms[i]
                other_atom = atoms[j]
                if i == j and np.linalg.norm(images[image_index]) <= EPS:
                    continue
                dist = distances[index]
                pair_key = tuple(sorted([atom.species, other_atom.species]))
                if pair_key not in element_bonds or element_bonds[pair_key] > dist:
                    element_bonds[pair_key] = dist
                # find the closest image of an atom to i
                if dist < min_dists[i]:
                    min_dists[i] = dist

            if debug:
                print(min_dists)
                print(element_bonds)

            for index in np.where(~distances.mask)[0]:
                image_index = int(index / num_atoms**2)
                i = int((index - image_index * num_atoms**2) / num_atoms)
                j = int((index - image_index * num_atoms**2) % num_atoms)
                atom = atoms[i]
                other_atom = atoms[j]
                min_dist = min_dists[i]
                if i == j and np.linalg.norm(images[image_index]) <= EPS:
                    continue
                dist = distances[index]
                pair_key = tuple(sorted([atom.species, other_atom.species]))
                if (
                    dist <= min_dist * coordination_cutoff
                    and dist <= element_bonds[pair_key] * bond_tolerance
                ):
                    if separate_images and all(
                        [val <= 0 + 1e-8 for val in images[image_index]]
                    ):
                        image_number += 1
                        self.add_node(j + image_number, species=atoms[j].species)
                        self.add_edge(i, j + image_number, dist=dist)
                        self.add_edge(j + image_number, i, dist=dist)
                    else:
                        is_image = np.linalg.norm(images[image_index]) > EPS
                        self.add_edge(i, j, dist=dist, image=is_image)

        elif graph is not None:
            for node, data in graph.nodes.data():
                self.add_node(
                    node, species=data["species"], image=data.get("image", False)
                )
            for node_in, node_out, data in graph.edges.data():
                if not delete_one_way_bonds or (node_out, node_in) in graph.edges():
                    self.add_edge(
                        node_in,
                        node_out,
                        dist=data.get("dist", 0),
                        image=data.get("image", False),
                    )

    def get_strongly_connected_component_subgraphs(self, delete_one_way_bonds=True):
        """Return generator of strongly-connected subgraphs in CrystalGraph format."""
        return (
            CrystalGraph(
                graph=self.subgraph(c), delete_one_way_bonds=delete_one_way_bonds
            )
            for c in nx.strongly_connected_components(self)
        )

    def get_communities(self, graph=None, **louvain_kwargs):
        """Return list of community subgraphs in CrystalGraph format."""
        import community as louvain

        if graph is None:
            graph = self

        if graph.is_directed():
            undirected_graph = self.remove_directionality(graph=graph)

        partition = louvain.best_partition(undirected_graph, **louvain_kwargs)
        size = len(set(partition.values()))
        subgraphs = [nx.MultiGraph() for i in range(size)]

        for node in partition:
            subgraphs[partition[node]].add_node(
                node,
                species=list(self.nodes(data=True))[list(self.nodes()).index(node)][1][
                    "species"
                ],
            )
        for edge in self.edges():
            if partition[edge[0]] == partition[edge[1]]:
                subgraphs[partition[edge[0]]].add_edge(edge[0], edge[1])

        subgraphs = [CrystalGraph(graph=sg) for sg in subgraphs]

        return subgraphs, partition

    def remove_directionality(self, graph=None):
        if graph is None:
            graph = self
        import networkx as nx

        undirected_graph = nx.MultiGraph()
        for node in graph.nodes(data=True):
            undirected_graph.add_node(node[0], species=node[1]["species"])
        for edge in graph.edges(data=True):
            if (edge[1], edge[0]) not in undirected_graph.edges():
                undirected_graph.add_edge(
                    edge[0],
                    edge[1],
                    dist=edge[2].get("dist", 0),
                    image=edge[2].get("image", False),
                )

        return undirected_graph

    def set_unique_subgraphs(self, method="community"):
        """Filter strongly connected component subgraphs for isomorphism with others inside
        CrystalGraph. Sets self.unique_subgraph to a set of such subgraphs.
        """
        if method == "community":
            self.unique_subgraphs = get_unique_subgraphs(self.get_communities())
        elif method == "strongly_connected":
            self.unique_subgraphs = get_unique_subgraphs(
                self.get_strongly_connected_component_subgraphs()
            )
        elif method == "both":
            strong_subgraphs = self.get_strongly_connected_component_subgraphs()
            community_subgraphs = []
            for sg in strong_subgraphs:
                community_subgraphs.extend(sg.get_communities())

            self.unique_subgraphs = get_unique_subgraphs(community_subgraphs)

    def get_bonds_per_atom(self):
        num_bonds = 0
        for node_in in self.nodes():
            for node_out in self.nodes():
                if node_in == node_out:
                    continue
                if self.has_edge(node_in, node_out) and self.has_edge(
                    node_out, node_in
                ):
                    num_bonds += 1
        return num_bonds / self.number_of_nodes()


def node_match(n1, n2):
    return n1["species"] == n2["species"]


def get_unique_subgraphs(subgraphs):
    """Filter strongly connected component subgraphs for isomorphism with others.

    Input:

        | subgraphs: list(CrystalGraph), list of subgraph objects to filter

    Returns:

        | unique_subgraphs: set(CrystalGraph), set of unique subgraphs

    """
    unique_subgraphs = set()
    for subgraph in subgraphs:
        if not any(
            [
                are_graphs_the_same(subgraph, other_subgraph)
                for other_subgraph in unique_subgraphs
            ]
        ):
            unique_subgraphs.add(subgraph)
    return unique_subgraphs


def are_graphs_the_same(g1, g2, edge_match=None):
    if edge_match is None:

        def edge_match(e1, e2):
            atol = 0.1
            rtol = 0.05
            return (
                abs(e1[0]["dist"] - e2[0]["dist"]) <= atol + rtol * e2[0]["dist"]
                and e1[0]["image"] == e2[0]["image"]
            )

    return nx.is_isomorphic(
        g1,
        g2,
        node_match=lambda n1, n2: n1["species"] == n2["species"],
        edge_match=edge_match,
    )


def draw_network(
    structure,
    layout=None,
    edge_labels=False,
    node_index=False,
    curved_edges=True,
    node_colour="elem",
    partition=None,
    ax=None,
):
    import networkx as nx
    from matador.utils.viz_utils import get_element_colours
    import matplotlib.pyplot as plt

    element_colours = get_element_colours()
    try:
        network = structure.network
    except Exception:
        network = structure
    if layout is None:
        pos = nx.spring_layout(network)
    else:
        pos = layout

    if ax is None:
        fig, ax = plt.subplots()

    if node_colour == "degree":
        coords = list(set(dict(network.degree).values()))
        cmap = plt.cm.get_cmap("Dark2", len(coords)).colors
        colours = [cmap[coords.index(network.degree[node])] for node in network.nodes()]

    elif node_colour == "partition" and partition is not None:
        num_partitions = len(set(partition.values()))
        cmap = plt.cm.get_cmap("Dark2", num_partitions).colors
        colours = [cmap[partition[node]] for node in network.nodes()]

    else:
        elem_map = element_colours
        colours = [elem_map.get(data["species"]) for node, data in network.nodes.data()]

    if node_index:
        labels = {
            node: "{} \\#{}".format(data["species"], node)
            for node, data in network.nodes.data()
        }
    else:
        labels = {node: str(data["species"]) for node, data in network.nodes.data()}

    edge_colours = []
    for edge in network.edges(data=True):
        if edge[2].get("image", True):
            edge_colours.append("grey")
        else:
            edge_colours.append("black")

    nx.draw_networkx_nodes(
        network,
        pos,
        node_color=colours,
        edgecolors="black",
        linewidths=2,
        node_size=1000,
        ax=ax,
    )
    nx.draw_networkx_edges(
        network, pos, edge_color=edge_colours, width=2, node_size=1000, ax=ax
    )
    if edge_labels:
        edge_weight = dict()
        for edge in network.edges(data=True):
            # data = edge[2]
            edge = (edge[0], edge[1])
            if edge not in edge_weight and (edge[1], edge[0]) not in edge_weight:
                edge_weight[edge] = 1
            else:
                if edge in edge_weight:
                    edge_weight[edge] += 1
                else:
                    edge_weight[(edge[1], edge[0])] += 1
        edge_label_dict = edge_weight
        nx.draw_networkx_edge_labels(network, pos, edge_labels=edge_label_dict, ax=ax)
    nx.draw_networkx_labels(network, pos, labels=labels, ax=ax)
    plt.axis("off")



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The query module provides the DBQuery class that implements all
database queries, either for individual compositions/formulae, or for
matching calculations at the same accuracy to create phase diagrams.

"""


__all__ = ["DBQuery"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .query import DBQuery



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the base PhaseDiagram creator that interfaces
with QueryConvexHull and EnsembleHull.

"""


from traceback import print_exc
import bisect

import scipy.spatial
import numpy as np

from matador.utils.hull_utils import (
    barycentric2cart,
    vertices2plane,
    vertices2line,
    FakeHull,
    is_point_in_triangle,
)
from matador.utils.chem_utils import get_formula_from_stoich
from matador.utils.cursor_utils import (
    get_array_from_cursor,
    display_results,
    set_cursor_from_array,
)

EPS = 1e-12


class PhaseDiagram:
    """This class encapsulates the actual phase data, e.g. the actual
    energy and compositions found to be stable.

    Attributes:
        structures (numpy.ndarray): the array passed to init used to
            make the hull, with the first (num_species-1) columns
            containing normalised concentrations, and the final column
            containing formation energy.
        convex_hull (scipy.spatial.ConvexHull): the actual convex hull
            returned by SciPy.
        formation_key (list): index/key specification of formation energy
            per atom from top level of each document.

    """

    def __init__(self, cursor, formation_key, dimension):
        """Compute the convex hull of data passed, to retrieve hull
        distances and thus stable structures.

        Parameters:
            cursor (list[dict]): list of matador documents to make
                phase diagram from.
            formation_key (str or list):  location of the formation energy
                inside each document, either a single key or iterable of
                keys to use with `recursive_get`.

        """
        self._dimension = dimension
        self.cursor = cursor
        self.formation_key = formation_key

        structures = np.hstack(
            (
                get_array_from_cursor(cursor, "concentration").reshape(
                    len(cursor), dimension - 1
                ),
                get_array_from_cursor(cursor, self.formation_key).reshape(
                    len(cursor), 1
                ),
            )
        )

        # define self._structure_slice as the filtered array of points actually used to create the convex hull
        # which can include/exclude points from the passed structures. This array is the one indexed by
        # vertices/simplices in ConvexHull

        if self._dimension == 3:
            # add a point "above" the hull
            # for simple removal of extraneous vertices (e.g. top of 2D hull)
            dummy_point = [0.333, 0.333, 1e5]
            # if ternary, use all structures, not just those with negative eform for compatibility reasons
            self._structure_slice = np.vstack((structures, dummy_point))
        else:
            # filter out those with positive formation energy, to reduce expense computing hull
            self._structure_slice = structures[np.where(structures[:, -1] <= 0 + EPS)]

        # filter out "duplicates" in _structure_slice
        # this prevents breakages if no structures are on the hull and chempots are duplicated
        # but it might be faster to hardcode this case individually
        self._structure_slice = np.unique(self._structure_slice, axis=0)

        # if we only have the chempots (or worse) with negative formation energy, don't even make the hull
        if len(self._structure_slice) <= dimension:
            if len(self._structure_slice) < dimension:
                raise RuntimeError(
                    "No chemical potentials on hull... either mysterious use of custom chempots, or worry!"
                )
            self.convex_hull = FakeHull()
        else:
            try:
                self.convex_hull = scipy.spatial.ConvexHull(self._structure_slice)
            except scipy.spatial.qhull.QhullError:
                print(self._structure_slice)
                print("Error with QHull, plotting formation energies only...")
                print_exc()
                self.convex_hull = FakeHull()

        # remove vertices that have positive formation energy
        filtered_vertices = [
            vertex
            for vertex in self.convex_hull.vertices
            if self._structure_slice[vertex, -1] <= 0 + EPS
        ]
        bad_simplices = set()
        for ind, simplex in enumerate(self.convex_hull.simplices):
            for vertex in simplex:
                if vertex not in filtered_vertices:
                    bad_simplices.add(ind)

        filtered_simplices = [
            simplex
            for ind, simplex in enumerate(self.convex_hull.simplices)
            if ind not in bad_simplices
        ]

        self.convex_hull = FakeHull()
        self.convex_hull.points = self._structure_slice
        self.convex_hull.vertices = list(filtered_vertices)
        self.convex_hull.simplices = list(filtered_simplices)

        self.hull_dist = self.get_hull_distances(structures, precompute=True)
        set_cursor_from_array(self.cursor, self.hull_dist, "hull_distance")
        self.structures = structures
        self.stable_structures = [
            doc for doc in self.cursor if doc["hull_distance"] < EPS
        ]

    def __str__(self):
        """Print underlying phase diagram."""
        return display_results(
            self.cursor,
            hull=True,
            colour=False,
            energy_key=self.formation_key,
            sort=False,
            return_str=True,
        )

    def get_hull_distances(self, structures, precompute=False, **kwargs):
        """Returns array of distances to pre-computed binary or ternary
        hull, from array containing concentrations and energies.

        Parameters:
            structures (numpy.ndarray): N x n array of concentrations and
                enthalpies for N structures, with up to 2 columns of
                concentrations and the last column containing the
                structure's formation enthalpy.

        Keyword arguments:
            precompute (bool): whether or not to bootstrap hull
                distances from previously computed values at the same
                stoichiometry.

        Returns:
            numpy.ndarray: N-dim array storing distances to
                the hull for N structures,

        """

        if precompute:
            # dict with formula keys, containing tuple of pre-computed enthalpy/atom and hull distance
            cached_formula_dists = dict()
            cache_hits = 0
            cache_misses = 0

        if isinstance(structures, list):
            structures = np.asarray(structures)

        # if only chem pots on hull, dist = energy
        if len(self._structure_slice) == self._dimension:
            hull_dist = np.ones((len(structures)))
            hull_dist = structures[:, -1]

        # if binary hull, do binary search
        elif self._dimension == 2:
            tie_line_comp = self._structure_slice[self.convex_hull.vertices, 0]
            tie_line_energy = self._structure_slice[self.convex_hull.vertices, -1]
            tie_line_comp = np.asarray(tie_line_comp)
            tie_line_energy = tie_line_energy[np.argsort(tie_line_comp)]
            tie_line_comp = tie_line_comp[np.argsort(tie_line_comp)]

            hull_dist = np.empty((len(structures)))
            hull_dist.fill(np.nan)
            if precompute:
                for ind, _ in enumerate(structures):
                    formula = get_formula_from_stoich(
                        self.cursor[ind]["stoichiometry"], sort=True, tex=False
                    )
                    if formula in cached_formula_dists:
                        hull_dist[ind] = (
                            structures[ind, -1]
                            - cached_formula_dists[formula][0]
                            + cached_formula_dists[formula][1]
                        )
                        cache_hits += 1
                    else:
                        i = bisect.bisect_left(tie_line_comp, structures[ind, 0])
                        gradient, intercept = vertices2line(
                            [
                                [tie_line_comp[i - 1], tie_line_energy[i - 1]],
                                [tie_line_comp[i], tie_line_energy[i]],
                            ]
                        )
                        # calculate hull_dist
                        hull_dist[ind] = structures[ind, -1] - (
                            gradient * structures[ind, 0] + intercept
                        )
                        cached_formula_dists[formula] = (
                            structures[ind, -1],
                            hull_dist[ind],
                        )
                        cache_misses += 1
            else:
                for ind, _ in enumerate(structures):
                    i = bisect.bisect_left(tie_line_comp, structures[ind, 0])
                    gradient, intercept = vertices2line(
                        [
                            [tie_line_comp[i - 1], tie_line_energy[i - 1]],
                            [tie_line_comp[i], tie_line_energy[i]],
                        ]
                    )
                    # calculate hull_dist
                    hull_dist[ind] = structures[ind, -1] - (
                        gradient * structures[ind, 0] + intercept
                    )

        # if ternary, use barycentric coords
        elif self._dimension == 3:
            # loop through structures and find which plane they correspond to
            # using barycentric coordinates, if a formula has already been
            # computed then calculate delta relative to that and skip
            self.convex_hull.planes = [
                [self._structure_slice[vertex] for vertex in simplex]
                for simplex in self.convex_hull.simplices
            ]
            structures_finished = [False] * len(structures)
            hull_dist = np.empty(len(structures))
            hull_dist.fill(np.nan)
            cart_planes_inv = []
            planes_height_fn = []
            for ind, plane in enumerate(self.convex_hull.planes):
                cart_planes = barycentric2cart(plane).T
                cart_planes[-1, :] = 1
                # if projection of triangle in 2D is a line, do binary search
                if np.linalg.det(cart_planes) == 0:
                    cart_planes_inv.append(None)
                    planes_height_fn.append(None)
                else:
                    cart_planes_inv.append(np.linalg.inv(cart_planes))
                    planes_height_fn.append(vertices2plane(plane))
            for idx, structure in enumerate(structures):
                for ind, plane in enumerate(self.convex_hull.planes):
                    if cart_planes_inv[ind] is None:
                        continue
                    if (
                        precompute
                        and get_formula_from_stoich(
                            self.cursor[idx]["stoichiometry"], sort=True, tex=False
                        )
                        in cached_formula_dists
                    ):
                        formula = get_formula_from_stoich(
                            self.cursor[idx]["stoichiometry"], sort=True, tex=False
                        )
                        if formula in cached_formula_dists:
                            cache_hits += 1
                            hull_dist[idx] = (
                                structures[idx, -1]
                                - cached_formula_dists[formula][0]
                                + cached_formula_dists[formula][1]
                            )
                            structures_finished[idx] = True

                    elif is_point_in_triangle(
                        structure, cart_planes_inv[ind], preprocessed_triangle=True
                    ):
                        structures_finished[idx] = True
                        hull_dist[idx] = planes_height_fn[ind](structure)
                        if precompute:
                            cached_formula_dists[
                                get_formula_from_stoich(
                                    self.cursor[idx]["stoichiometry"],
                                    sort=True,
                                    tex=False,
                                )
                            ] = (structure[-1], hull_dist[idx])
                            cache_misses += 1
                        break

            # mask values very close to 0 with 0
            hull_dist[np.where(np.abs(hull_dist) < EPS)] = 0

            failed_structures = []
            for ind, structure in enumerate(structures_finished):
                if not structure:
                    failed_structures.append(ind)

            if failed_structures:
                raise RuntimeError(
                    "There were issues calculating the hull distance for {} structures.".format(
                        len(failed_structures)
                    )
                )

        # otherwise, set to zero until proper N-d distance can be implemented
        else:
            raise NotImplementedError(
                "Unable to compute {dimension}-dimensional hull distances (yet) "
                "consider breaking your phase diagram into a pseudo-ternary or pseudo-binary system."
            )

        if np.isnan(hull_dist).any():
            raise RuntimeError(
                f"Some hull distances failed, found NaNs at {np.isnan(hull_dist, where=True)}"
            )

        return hull_dist



from pathlib import Path

with open(Path(__file__).parent.parent.joinpath("data/words").resolve(), "r") as f:
    WORDS = f.readlines()
with open(Path(__file__).parent.parent.joinpath("data/nouns").resolve(), "r") as f:
    NOUNS = f.readlines()

__all__ = ["NOUNS", "WORDS"]



"""This submodule implements some convenience dictionaries for working with the periodic table.

Data was scraped from a combination of the periodictable Python package (on which matador used to
depend) and pymatgen.

"""
from typing import Dict
from dataclasses import dataclass

__all__ = ("PERIODIC_TABLE",)


@dataclass
class Element:
    symbol: str
    number: int
    mass: float
    iupac_order: int


PERIODIC_TABLE: Dict[str, Element] = {
    "n": Element(symbol="n", number=0, mass=0, iupac_order=-1),
    "H": Element(symbol="H", number=1, mass=1.00794, iupac_order=92),
    "He": Element(symbol="He", number=2, mass=4.002602, iupac_order=5),
    "Li": Element(symbol="Li", number=3, mass=6.941, iupac_order=11),
    "Be": Element(symbol="Be", number=4, mass=9.012182, iupac_order=17),
    "B": Element(symbol="B", number=5, mass=10.811, iupac_order=81),
    "C": Element(symbol="C", number=6, mass=12.0107, iupac_order=86),
    "N": Element(symbol="N", number=7, mass=14.0067, iupac_order=91),
    "O": Element(symbol="O", number=8, mass=15.9994, iupac_order=97),
    "F": Element(symbol="F", number=9, mass=18.9984032, iupac_order=102),
    "Ne": Element(symbol="Ne", number=10, mass=20.1797, iupac_order=4),
    "Na": Element(symbol="Na", number=11, mass=22.98977, iupac_order=10),
    "Mg": Element(symbol="Mg", number=12, mass=24.305, iupac_order=16),
    "Al": Element(symbol="Al", number=13, mass=26.981538, iupac_order=80),
    "Si": Element(symbol="Si", number=14, mass=28.0855, iupac_order=85),
    "P": Element(symbol="P", number=15, mass=30.973761, iupac_order=90),
    "S": Element(symbol="S", number=16, mass=32.065, iupac_order=96),
    "Cl": Element(symbol="Cl", number=17, mass=35.453, iupac_order=101),
    "Ar": Element(symbol="Ar", number=18, mass=39.948, iupac_order=3),
    "K": Element(symbol="K", number=19, mass=39.0983, iupac_order=9),
    "Ca": Element(symbol="Ca", number=20, mass=40.078, iupac_order=15),
    "Sc": Element(symbol="Sc", number=21, mass=44.95591, iupac_order=49),
    "Ti": Element(symbol="Ti", number=22, mass=47.867, iupac_order=52),
    "V": Element(symbol="V", number=23, mass=50.9415, iupac_order=55),
    "Cr": Element(symbol="Cr", number=24, mass=51.9961, iupac_order=58),
    "Mn": Element(symbol="Mn", number=25, mass=54.938049, iupac_order=61),
    "Fe": Element(symbol="Fe", number=26, mass=55.845, iupac_order=64),
    "Co": Element(symbol="Co", number=27, mass=58.9332, iupac_order=67),
    "Ni": Element(symbol="Ni", number=28, mass=58.6934, iupac_order=70),
    "Cu": Element(symbol="Cu", number=29, mass=63.546, iupac_order=73),
    "Zn": Element(symbol="Zn", number=30, mass=65.409, iupac_order=76),
    "Ga": Element(symbol="Ga", number=31, mass=69.723, iupac_order=79),
    "Ge": Element(symbol="Ge", number=32, mass=72.64, iupac_order=84),
    "As": Element(symbol="As", number=33, mass=74.9216, iupac_order=89),
    "Se": Element(symbol="Se", number=34, mass=78.96, iupac_order=95),
    "Br": Element(symbol="Br", number=35, mass=79.904, iupac_order=100),
    "Kr": Element(symbol="Kr", number=36, mass=83.798, iupac_order=2),
    "Rb": Element(symbol="Rb", number=37, mass=85.4678, iupac_order=8),
    "Sr": Element(symbol="Sr", number=38, mass=87.62, iupac_order=14),
    "Y": Element(symbol="Y", number=39, mass=88.90585, iupac_order=48),
    "Zr": Element(symbol="Zr", number=40, mass=91.224, iupac_order=51),
    "Nb": Element(symbol="Nb", number=41, mass=92.90638, iupac_order=54),
    "Mo": Element(symbol="Mo", number=42, mass=95.94, iupac_order=57),
    "Tc": Element(symbol="Tc", number=43, mass=98, iupac_order=60),
    "Ru": Element(symbol="Ru", number=44, mass=101.07, iupac_order=63),
    "Rh": Element(symbol="Rh", number=45, mass=102.9055, iupac_order=66),
    "Pd": Element(symbol="Pd", number=46, mass=106.42, iupac_order=69),
    "Ag": Element(symbol="Ag", number=47, mass=107.8682, iupac_order=72),
    "Cd": Element(symbol="Cd", number=48, mass=112.411, iupac_order=75),
    "In": Element(symbol="In", number=49, mass=114.818, iupac_order=78),
    "Sn": Element(symbol="Sn", number=50, mass=118.71, iupac_order=83),
    "Sb": Element(symbol="Sb", number=51, mass=121.76, iupac_order=88),
    "Te": Element(symbol="Te", number=52, mass=127.6, iupac_order=94),
    "I": Element(symbol="I", number=53, mass=126.90447, iupac_order=99),
    "Xe": Element(symbol="Xe", number=54, mass=131.293, iupac_order=1),
    "Cs": Element(symbol="Cs", number=55, mass=132.90545, iupac_order=7),
    "Ba": Element(symbol="Ba", number=56, mass=137.327, iupac_order=13),
    "La": Element(symbol="La", number=57, mass=138.9055, iupac_order=47),
    "Ce": Element(symbol="Ce", number=58, mass=140.116, iupac_order=46),
    "Pr": Element(symbol="Pr", number=59, mass=140.90765, iupac_order=45),
    "Nd": Element(symbol="Nd", number=60, mass=144.24, iupac_order=44),
    "Pm": Element(symbol="Pm", number=61, mass=145, iupac_order=43),
    "Sm": Element(symbol="Sm", number=62, mass=150.36, iupac_order=42),
    "Eu": Element(symbol="Eu", number=63, mass=151.964, iupac_order=41),
    "Gd": Element(symbol="Gd", number=64, mass=157.25, iupac_order=40),
    "Tb": Element(symbol="Tb", number=65, mass=158.92534, iupac_order=39),
    "Dy": Element(symbol="Dy", number=66, mass=162.5, iupac_order=38),
    "Ho": Element(symbol="Ho", number=67, mass=164.93032, iupac_order=37),
    "Er": Element(symbol="Er", number=68, mass=167.259, iupac_order=36),
    "Tm": Element(symbol="Tm", number=69, mass=168.93421, iupac_order=35),
    "Yb": Element(symbol="Yb", number=70, mass=173.04, iupac_order=34),
    "Lu": Element(symbol="Lu", number=71, mass=174.967, iupac_order=33),
    "Hf": Element(symbol="Hf", number=72, mass=178.49, iupac_order=50),
    "Ta": Element(symbol="Ta", number=73, mass=180.9479, iupac_order=53),
    "W": Element(symbol="W", number=74, mass=183.84, iupac_order=56),
    "Re": Element(symbol="Re", number=75, mass=186.207, iupac_order=59),
    "Os": Element(symbol="Os", number=76, mass=190.23, iupac_order=62),
    "Ir": Element(symbol="Ir", number=77, mass=192.217, iupac_order=65),
    "Pt": Element(symbol="Pt", number=78, mass=195.078, iupac_order=68),
    "Au": Element(symbol="Au", number=79, mass=196.96655, iupac_order=71),
    "Hg": Element(symbol="Hg", number=80, mass=200.59, iupac_order=74),
    "Tl": Element(symbol="Tl", number=81, mass=204.3833, iupac_order=77),
    "Pb": Element(symbol="Pb", number=82, mass=207.2, iupac_order=82),
    "Bi": Element(symbol="Bi", number=83, mass=208.98038, iupac_order=87),
    "Po": Element(symbol="Po", number=84, mass=209, iupac_order=93),
    "At": Element(symbol="At", number=85, mass=210, iupac_order=98),
    "Rn": Element(symbol="Rn", number=86, mass=222, iupac_order=0),
    "Fr": Element(symbol="Fr", number=87, mass=223, iupac_order=6),
    "Ra": Element(symbol="Ra", number=88, mass=226, iupac_order=12),
    "Ac": Element(symbol="Ac", number=89, mass=227, iupac_order=32),
    "Th": Element(symbol="Th", number=90, mass=232.0381, iupac_order=31),
    "Pa": Element(symbol="Pa", number=91, mass=231.03588, iupac_order=30),
    "U": Element(symbol="U", number=92, mass=238.02891, iupac_order=29),
}



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the AtomicSwapper class which
takes a Query object and a desired "swap", e.g. swap all group [I]
elements to Li, ready to be re-relaxed.

"""


__all__ = ["AtomicSwapper"]
__author__ = "Matthew Evans"
__maintainer__ = "Matthew Evans"


from .swaps import AtomicSwapper



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the :class:`CastepMagresWorkflow`
class, which performs magres calculations with CASTEP in
multiple steps (only when necessary):

    1. Try to pre-relax structure (skipped if check file
       is already present).
    2. Perform an SCF with lower electronic tolerances.
    3. Calculate NMR properties, e.g. shielding and EFG.

"""


import copy
import logging
import os
from functools import partial

from matador.scrapers import arbitrary2dict
from matador.workflows.castep.common import castep_prerelax, castep_scf
from matador.workflows.castep.spectral import (
    _get_optados_fname,
    castep_spectral_dos,
    optados_dos_broadening,
    optados_pdos,
)
from matador.workflows.workflows import Workflow

LOG = logging.getLogger("run3")

__all__ = "castep_full_magres"


def castep_full_magres(computer, calc_doc, seed, final_elec_energy_tol=1e-11, **kwargs):
    """Perform a "full" magres calculation on a system, i.e.
    first perform a relaxation, then do a high quality SCF
    and compute NMR properties in the same step.

    This function is a wrapper for the :class:`CastepMagresWorkflow` class.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): dictionary of structure and calculation
            parameters.
        seed (str): root seed for the calculation.

    Raises:
        RuntimeError: if any part of the calculation fails.

    Returns:
        bool: True if Workflow completed successfully, or False otherwise.

    """
    workflow = CastepMagresWorkflow(
        computer, calc_doc, seed, final_elec_energy_tol=final_elec_energy_tol, **kwargs
    )

    return workflow.success


class CastepMagresWorkflow(Workflow):
    """Perform a "full" magres calculation on a system, i.e.
    first perform a relaxation in a standardised unit cell,
    then do a high quality SCF, then compute NMR properties.

    Attributes:
        computer (:obj:`ComputeTask`): the object that calls CASTEP.
        calc_doc (dict): the interim dictionary of structural and
            calculation parameters.
        seed (str): the root seed for the calculation.
        success (bool): the status of the Workflow: only set to True after
            post-processing method completes.
        final_elec_energy_tol (float): the electronic energy tolerance to use
            in the high-quality SCF calculation.

    """

    def preprocess(self):
        """Decide which parts of the Workflow need to be performed,
        and set the appropriate CASTEP parameters.

        """

        self.final_elec_energy_tol = self.workflow_params.get(
            "final_elec_energy_tol", 1e-11
        )
        # default todo
        todo = {
            "relax": True,
            "scf": True,
            "dos": True,
            "pdos": True,
            "broadening": True,
            "magres": True,
        }
        # definition of steps and names
        steps = {
            "relax": castep_prerelax,
            "scf": partial(
                castep_magres_scf, elec_energy_tol=self.final_elec_energy_tol
            ),
            "dos": castep_spectral_dos,
            "pdos": optados_pdos,
            "broadening": optados_dos_broadening,
            "magres": partial(
                castep_magres, elec_energy_tol=self.final_elec_energy_tol
            ),
        }

        exts = {
            "relax": {
                "input": [".cell", ".param"],
                "output": [".castep", "-out.cell", ".*err"],
            },
            "scf": {"input": [".cell", ".param"], "output": [".castep", ".bands"]},
            "magres": {"input": [".cell", ".param"], "output": [".castep", ".magres"]},
            "dos": {
                "input": [".cell", ".param"],
                "output": [
                    ".castep",
                    ".bands",
                    ".pdos_bin",
                    ".dome_bin",
                    ".*err",
                    "-out.cell",
                ],
            },
            "pdos": {
                "input": [".odi", ".pdos_bin", ".dome_bin"],
                "output": [".odo", ".*err"],
            },
            "broadening": {
                "input": [".odi", ".pdos_bin", ".dome_bin"],
                "output": [".odo", ".*err"],
            },
        }

        odi_fname = _get_optados_fname(self.seed)
        if odi_fname is not None:
            odi_dict, _ = arbitrary2dict(odi_fname)
            if todo["dos"]:
                todo["broadening"] = "broadening" in odi_dict
                todo["pdos"] = "pdos" in odi_dict
        else:
            todo["dos"] = False
            todo["pdos"] = False
            todo["broadening"] = False

        # prepare to do pre-relax if there's no check file
        if os.path.isfile(self.seed + ".check"):
            todo["scf"] = True
            todo["relax"] = False
            LOG.info(
                "Restarting from {}.check, so not performing re-relaxation".format(
                    self.seed
                )
            )

        # If geom force tol is not set, do not perform a relaxation
        if self.calc_doc.get("geom_force_tol") is None:
            todo["relax"] = False

        for key in todo:
            if todo[key]:
                self.add_step(
                    steps[key],
                    key,
                    input_exts=exts[key].get("input"),
                    output_exts=exts[key].get("output"),
                )


def castep_magres_scf(computer, calc_doc, seed, elec_energy_tol=1e-11):
    """Run a singleshot SCF calculation with a high elec_energy_tol.

    Parameters:
        computer (:obj:`matador.compute.ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    Returns:
        bool: whether or not the SCF was successful.

    """
    calc_doc["write_checkpoint"] = "ALL"
    calc_doc["continuation"] = "default"

    required = ["write_checkpoint", "continuation"]

    return castep_scf(
        computer,
        calc_doc,
        seed,
        elec_energy_tol=elec_energy_tol,
        required_keys=required,
    )


def castep_magres(computer, calc_doc, seed, elec_energy_tol=1e-11):
    """Runs a NMR properties calculation on top of a completed
    SCF calculation.

    Parameters:
        computer (:obj:`ComputeTask`): the object that will be calling CASTEP.
        calc_doc (dict): the structure to run on.
        seed (str): root filename of structure.

    """
    LOG.info("Performing CASTEP Magres calculation...")
    magres_doc = copy.deepcopy(calc_doc)
    magres_doc["task"] = "magres"
    magres_doc["magres_task"] = calc_doc.get("magres_task", "NMR")
    magres_doc["continuation"] = "default"
    # this is just to suppress a warning that elec_energy_tol has changed
    magres_doc["elec_energy_tol"] = elec_energy_tol

    required = []
    forbidden = []

    computer.validate_calc_doc(magres_doc, required, forbidden)

    return computer.run_castep_singleshot(
        magres_doc, seed, keep=True, intermediate=True
    )
